# -*- coding: utf-8 -*- vim: ts=8 sts=4 sw=4 si et tw=79
"""
visaplan.plone.tools.lock: Some sugar for zc.lockfile.LockFile

We depend on zc.lockfile in this module;
since you might not need it, we don't actually require that package in our metadata.

We add some sugar to the zc.lockfile.LockFile class, e.g. a context manager::

  form zc.lockfile import LockError
  from visaplan.plone.tools.lock import lock
  lfn = '/var/lock/perfectly-normal.lock'
  try:
      with lock(lfn):
          # do what you need to do ...
  except LockError as e:
      # by default, we won't catch that error

When leaving the context, we try to delete the lock file by default;
specify autodelete=False to prevent this.
We'll ignore any OSError which may occur during deletion, however;
if verbose (requires a logger to be specified), this will be logged.

Further remarks:

- The property ``filename`` provides the name used internally to create the
  zc.lockfile.LockFile;
- The property ``lockfile`` (not to be confused with the former) provides
  that very LockFile object.
- There is an ``acquire`` method you may use to get the lock,
  although the usual way will be to simply `__enter__` the context;
- there is a companion ``release`` method
  which is normally called by the ``__exit__`` method.

"""

# Python compatibility:
from __future__ import absolute_import, print_function

from importlib_metadata import PackageNotFoundError
from importlib_metadata import version as pkg_version

# Setup tools:
from packaging.version import parse as parse_version

# Standard library:
from os import unlink
from time import sleep

try:
    ZCLF_V = parse_version(pkg_version('zc.lockfile'))
except PackageNotFoundError:
    if __name__ == '__main__':
        ZCLF_V = parse_version('1.2.3.post4')
        print("Package zc.lockfile not found; some tests probably won't work!")
        class LockFile: pass
        class LockError(Exception): pass
    else:
        raise
else:
    # Zope:
    from zc.lockfile import LockError, LockFile

# Local imports:
from visaplan.plone.tools.env import worker_name

__all__ = [
    'ConvenientLock',  # our wrapper class, a context manager
    'lockfile_kwargs',  # build keyword args for zc.lockfile.LockFile creation
    ]

HAVE_CUSTOMCONTENT = ZCLF_V >= parse_version('1.2.0')  # 1.2.1+ recommended!
DEFAULTS_KW = {
    'add_pid': True,
    'add_hostname': False,
    'add_worker': False,
    'content_template': None,
    'sep': ';',
    }

def lockfile_kwargs(_zclf_v=None, _env=None, **kwargs):
    """
    Forge keyword arguments for zc.lockfile.LockFile;
    the first argument `name`, however, we expect to be specified positionally.

    Let's assume we have a recent zc.lockfile release,
    and we want the hostname to be included in the lockfile contents:

    >>> def lfkw_recent(**kw):
    ...     _zclf_v = parse_version('3.0.post1')
    ...     return lockfile_kwargs(_zclf_v, **kw)
    >>> lfkw_recent(add_hostname=1)
    {'content_template': '{pid};{hostname}'}

    If we have a pre-1.2.0 version, as pinned for Plone 4.3 versions by
    default, the content_template argument is not supported yet,
    so we'll get an empty dict:

    >>> def lfkw_ancient(**kw):
    ...     _zclf_v = parse_version('1.0.2')
    ...     return lockfile_kwargs(_zclf_v, **kw)
    >>> lfkw_ancient(add_hostname=1)
    {}

    In our Zope/Plone instances, we are sometimes interested in the "worker"
    (or part, in zc.buildout terms)
    which does certain things. We could look for the PID, of course,
    but the worker's name is much more "human readable";
    and since we just need to detect it once for each process ...

    We use the `.env.worker_name` for this purpose,
    which uses the environment, currently:

    >>> e = dict(INSTANCE_HOME='/opt/zope/parts/client3')
    >>> kw = {'add_worker': 1, '_env': e}
    >>> lfkw_recent(**kw)
    {'content_template': '{pid};client3'}

    Since there is no 'worker' support in zc.lockfile.LockFile,
    we simply use the value here.

    Now for some corner cases.

    Despite the fact that we "consume" the kwargs dict
    when processing the keyword arguments, it seems be be copied implicitly,
    so we can repeat our call and get the same result again:

    >>> lfkw_recent(**kw)
    {'content_template': '{pid};client3'}

    Arguments for the ConvenientLock class are not accepted here:
    >>> lfkw_recent(autodelete=1)
    Traceback (most recent call last):
      ...
    TypeError: Unsupported arguments!

    If the content_template is readily specified, we don't accept additional
    arguments, to avoid possible ambiguities.
    To avoid surprises, we fail for errors even if the zc.lockfile version
    doesn't support customized ocntent yet anyway:

    >>> lfkw_ancient(add_hostname=1, content_template='{pid}')
    Traceback (most recent call last):
      ...
    TypeError: With content_template given, no other arguments are allowed!

    """
    pop = kwargs.pop
    if _zclf_v is None:
        _zclf_v = ZCLF_V
    have_customcontent = _zclf_v > parse_version('1.2.0')

    _ct = pop('content_template', None)
    if _ct is not None:
        if kwargs:
            raise TypeError('With content_template given, no other arguments '
                            'are allowed!')
        elif have_customcontent:
            return {'content_template': _ct}
        else:
            return {}
    sep = pop('sep', ';')
    added = [key[4:] for key in DEFAULTS_KW.keys()
             if key.startswith('add_') and pop(key, DEFAULTS_KW[key])
             ]
    if kwargs:
        raise TypeError('Unsupported arguments!')
    if added == ['pid'] or not have_customcontent:
        # this is the default:
        return {}
    res = [key.join('{}') for key in ['pid', 'hostname']
           if key in added]
    if 'worker' in added:
        w = worker_name(_env)
        if w:
            res.append(w)
    return {'content_template': sep.join(res)}


class ConvenientLock(object):
    """
    A convience wrapper for zc.lockfile.LockFile
    """

    def __init__(self, name, **kwargs):
        self.__filename = name
        self.__lockfile = None
        pop = kwargs.pop
        self.__autodelete = pop('autodelete', 1)
        self._tries = tries = pop('tries', 1)
        if tries > 1:
            if 'delay' not in kwargs:
                raise TypeError('With tries=%(tries)r, we demand a delay value!'
                                % locals())
            self._delay = delay = pop('delay')
            if not isinstance(delay, (int, float)):
                raise ValueError('Number expected; found %s' % (type(delay),))
            elif delay <= 0:
                raise ValueError('Number [seconds] > 0 expected; found %r'
                                 % (delay,
                                    ))
        self._logger = logger = pop('logger', None)
        self._verbose = verbose = pop('verbose',
                              1 if logger is not None
                              else 0)
        if verbose and not logger:
            raise TypeError("Can't be verbose with no logger given!")
        self._lfkw = lockfile_kwargs(**kwargs)

    def __enter__(self):
        if not self.active:
            self.acquire()

    def acquire(self):
        i = 1
        tries = self._tries
        if tries > 1:
            delay = self._delay
        verbose = self._verbose
        if verbose:
            logger = self._logger
        filename = self.__filename
        while True:
            try:
                self.__lockfile = LockFile(filename, **self._lfkw)
            except LockError as e:
                if verbose:
                    msg = str(e)
                if i < tries:
                    i += 1
                    if verbose:
                        # (verified syntax:)
                        logger.warn('%s: sleep %f seconds, then retry ...',
                                    msg, delay)
                    sleep(delay)
                else:
                    if verbose:
                        logger.error(msg)
                    raise
            else:
                if verbose >= 2:
                    logger.info('Acquired lock: %r', self)
                break

    def __repr__(self):
        return "<%s('%s'): %s>" % (
                self.__class__.__name__,
                self.__filename,
                self.status,
                )

    @property
    def status(self):
        if self.active:
            return 'ACTIVE'
        elif self.__lockfile is None:
            return 'UNINITIALIZED'
        else:
            fp = self.__lockfile._fp
            if fp is None:
                return 'CLOSED'
            elif fp.closed:
                return 'closed'
            else:
                return '???'

    @property
    def lockfile(self):
        """
        This is the zc.lockfile.LockFile instance which we use internally
        """
        return self.__lockfile

    @property
    def filename(self):
        """
        This is the filename used to create the zc.lockfile.LockFile instance
        """
        return self.__filename

    @property
    def active(self):
        _lf = self.__lockfile
        if _lf is None:
            return False
        _fp = _lf._fp
        if _fp is None:
            return False
        return not _fp.closed

    def release(self):
        if self.active:
            self.__lockfile.close()
        verbose = self._verbose
        if verbose >= 1:
            logger = self._logger
        if self.__autodelete:
            filename = self.__filename
            try:
                unlink(filename)
            except OSError as e:
                if verbose:
                    logger.warn("LockFile %(filename)r NOT deleted!",
                                      locals())
            else:
                if verbose >= 2:
                    logger.info('Lockfile %(filename)r deleted!',
                                      locals())
        elif verbose >= 2:
            logger.info('Lockfile %r left behind.', self.__filename)
        if verbose >= 3:
            logger.info('Left context of %r', self)

    def __exit__(self, exc_type, exc_value, traceback):
        self.release()


if __name__ == '__main__':
    # Standard library:
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*- äöü vim: ts=8 sts=4 sw=4 si et hls tw=79
# Python compatibility:
from __future__ import absolute_import, print_function

from six import iteritems as six_iteritems
from six import text_type as six_text_type

# Zope:
from Products.CMFCore.utils import getToolByName

# visaplan:
from visaplan.tools.lands0 import list_of_strings

# Local imports:
from visaplan.plone.tools.context import make_translator

try:
    # visaplan:
    from visaplan.tools.coding import safe_decode
except ImportError:
    if __name__ != '__main__':
        raise
    def safe_decode(s):
        if isinstance(s, six_text_type):
            return s
        return s.decode('utf-8')




# -*- coding: utf-8 -*- äöü vim: ts=8 sts=4 sw=4 si et hls tw=79
# visaplan.plone.tools; groups package: information function factories

# Local imports:
from ._group import groupinfo_factory
from ._helpers import build_groups_set, split_group_id
from ._membership import (
    get_all_members,
    is_direct_member__factory,
    is_member_of__factory,
    is_member_of_any,
    recursive_members,
    )
from ._user import userinfo_factory

__all__ = [
    # _group.py:
    'groupinfo_factory',
    # _helpers.py:
    'build_groups_set', split_group_id,
    # _membership.py:
    'get_all_members',
    'is_direct_member__factory',
    'is_member_of__factory',
    'is_member_of_any',
    'recursive_members',
    # _user.py:
    'userinfo_factory',
    ]



# -*- coding: utf-8 -*- äöü vim: ts=8 sts=4 sw=4 si et hls tw=79
# Python compatibility:
from __future__ import absolute_import, print_function

from six import text_type as six_text_type

# Standard library:
from collections import defaultdict

# Zope:
from Products.CMFCore.utils import getToolByName

# visaplan:
from visaplan.tools.minifuncs import gimme_None

# Local imports:
from ._user_pio import _parse_init_options
from visaplan.plone.tools.context import make_translator

try:
    # visaplan:
    from visaplan.tools.coding import safe_decode
except ImportError:
    if __name__ != '__main__':
        raise
    def safe_decode(s):
        if isinstance(s, six_text_type):
            return s
        return s.decode('utf-8')


__all__ = [
    'userinfo_factory',
    ]


# --------------------------------- [ userinfo_factory ... [
def userinfo_factory(context, *args, **kwargs):
    """
    Wie groupinfo_factory, aber für Benutzerobjekte.

    pretty -- für Benutzer: title als formatierter Name
              (author.get_formatted_name), wenn möglich
    forlist -- Minimale Rückgabe (nur ID und Titel),
               aber mit pretty kombinierbar
    missing -- return a dict for missing users as well,
               and add an `exists` key

    title_or_id -- für Verwendung mit visaplan.tools.classes.Proxy:
               gib nur den Title oder ersatzweise die ID zurück
               (mit pretty kombinierbar)
    """
    _parse_init_options(kwargs, args)
    pretty = kwargs['pretty']
    forlist = kwargs['forlist']
    missing = kwargs['missing']
    missing_mask = kwargs.get('missing_user_mask') or None
    title_or_id = kwargs['title_or_id']

    acl = getToolByName(context, 'acl_users')
    acl_gu = acl.getUser
    if pretty or not forlist:
        author = context.restrictedTraverse('@@author', None)
        gbbuid = author.getBrainByUserId
        gfn = author.get_formatted_name
    if missing:
        assert missing_mask
        translate = make_translator(context)
        missing_mask = translate(missing_mask)

    # ---------------------------- [ forlist ... [
    def basic_user_info(member_id):
        """
        Basisinformationen über einen Benutzer:
        id, title

        forlist, not: pretty
        """
        member = acl_gu(member_id)
        if member:
            res = {
                'id': member_id,
                'title': member.getProperty('fullname'),
                }
            if missing:
                res['exists'] = 1
            return res
        elif missing:
            return defaultdict(gimme_None,
                               id=member_id,
                               exists=0)
        else:
            return {}

    def pretty_user_info(member_id):
        """
        Basisinformationen über einen Benutzer:
        id, title

        forlist, pretty
        """
        member = acl_gu(member_id)
        if member:
            brain = gbbuid(member_id)
            res = {
                'id': member_id,
                'title': (brain and gfn(brain))
                          or member.getProperty('fullname'),
                }
            if missing:
                res['exists'] = 1
        elif missing:
            res = defaultdict(gimme_None,
                              id=member_id,
                              exists=0)
            res['title'] = missing_mask.format(**res)
        else:
            return {}
        return res
    # ---------------------------- ] ... forlist ]

    # ------------------------ [ title_or_id ... [
    def pretty_title_or_id(member_id):
        dic = pretty_user_info(member_id)
        if dic or (missing and dic['exists']):
            return dic['title'] or member_id
        elif missing:
            return missing_mask.format(id=member_id)
        else:
            return None

    def basic_title_or_id(member_id):
        try:
            return basic_user_info(member_id)['title'] \
                   or member_id
        except:
            return None
    # ------------------------ ] ... title_or_id ]

    def full_user_info(member_id):
        """
        not: forlist, not: pretty
        """
        member = acl_gu(member_id)
        if member:
            brain = gbbuid(member_id)
            res = {
                'id': member_id,
                'title': member.getProperty('fullname'),
                'brain': brain,
                'email': member.getProperty('email'),
                }
            if missing:
                res['exists'] = 1
        elif missing:
            res = defaultdict(gimme_None,
                              id=member_id,
                              exists=0)
            res['title'] = missing_mask.format(**res)
        else:
            return {}
        return res

    def full_pretty_user_info(member_id):
        """
        not: forlist, pretty
        """
        member = acl_gu(member_id)
        if member:
            brain = gbbuid(member_id)
            res = {
                'id': member_id,
                'title': (brain and gfn(brain))
                         or member.getProperty('fullname'),
                'brain': brain,
                'email': member.getProperty('email'),
                }
            if missing:
                res['exists'] = 1
        elif missing:
            res = defaultdict(gimme_None,
                              id=member_id,
                              exists=0)
            res['title'] = missing_mask.format(**res)
        else:
            return {}
        return res

    if title_or_id:
        if pretty:
            return pretty_title_or_id
        else:
            return basic_title_or_id
    if forlist:
        if pretty:
            return pretty_user_info
        else:
            return basic_user_info
    if pretty:
        return full_pretty_user_info
    else:
        return full_user_info
# --------------------------------- ] ... userinfo_factory ]



# -*- coding: utf-8 -*- äöü vim: sw=4 sts=4 et tw=79
"""
Tools for Product Setup (GS migration steps, "upgrade steps")

We provide a few convenience tools here, to help you avoid to rebuild
*all* metadata and indexes whenever your setup has changed,
regardless of the nature of the changes.

The reindexing method of the catalog always expects an indexes specification --
or None, which means: all indexes. So, what will we do if we don't really care
about the indexes because all our changes apply to metadata only?
We'll specify a "cheap" subset of the indexes;
this subset is taken from the registry (if collective.metadataversion is
installed; see below), with a hardcoded fallback value ['getId'].

To go a step further and monitor the metadata, using a metadata_version
metadata attribute, you might want to use collective.metadataversion,
which as well allows you to adjust this subset.
Because of the slightly changed focus, the function there is named differently
as well: .utils.make_metadata_updater.
"""

# Python compatibility:
from __future__ import absolute_import, print_function

from importlib_metadata import PackageNotFoundError, version

try:
    version('collective.metadataversion')
except PackageNotFoundError:
    HAVE_METADATAVERSION = 0
else:
    HAVE_METADATAVERSION = 1

# Standard library:
from traceback import extract_stack

try:
    # Zope:
    import transaction
    from Products.CMFCore.utils import getToolByName
    from ZODB.POSException import POSKeyError

    # Local imports:
    from visaplan.plone.tools.setup._query import make_query_extractor
except ImportError:
    if __name__ != '__main__':  # doctests
        raise
    print('WARNING: Some imports failed; some tests might fail as well')

# Logging / Debugging:
import logging
from pdb import set_trace
from visaplan.tools.debug import pp

__all__ = [
        'make_reindexer',
        'reindex_all',       # calls make_reindexer internally
        'get_default_idxs',  # return a cheap subset
        ]


if HAVE_METADATAVERSION:
    # Zope:
    from zope.component import getUtility

    # Plone:
    from plone.registry.interfaces import IRegistry

    # 3rd party:
    from collective.metadataversion.config import DEFAULT_IDXS, FULL_IDXS_KEY

    def get_default_idxs():
        """
        We have collective.metadataversion
        """
        registry = getUtility(IRegistry)
        val = list(registry.get(FULL_IDXS_KEY) or DEFAULT_IDXS)
        assert val, 'The default list of indexes must not be falsy!'
        return val
else:
    def get_default_idxs():
        """
        We DON'T have collective.metadataversion
        """
        return ['getId']


def _update_mri_kwargs(logger, kwargs):
    """
    Update a kwargs dict, asserting the presence of ``update_metadata`` and
    ``idxs`` keys.

    NOTE: This function is for internal use, and both the signature and
          the functionality may change without notice!

    We use a little test helper function here:

    >>> def umri(kwargs):
    ...     _update_mri_kwargs(None, kwargs)
    ...     return sorted(kwargs.items())

    The tests also assume our default list of "cheap indexes":
    >>> get_default_idxs()
    ['getId']

    By default, we update the metadata; unless explicitly specified,
    the indexes specification defaults to that cheap subset:
    >>> kw = dict()
    >>> umri(kw)
    [('idxs', ['getId']), ('update_metadata', 1)]
    >>> kw = dict(idxs=None)
    >>> umri(kw)
    [('idxs', None), ('update_metadata', 1)]
    >>> kw = dict(idxs='idx1 idx2'.split())
    >>> umri(kw)
    [('idxs', ['idx1', 'idx2']), ('update_metadata', 1)]

    However, you can always choose yourself whether to update the metadata,
    of course:
    >>> kw = dict(idxs=None, update_metadata=0)
    >>> umri(kw)
    [('idxs', None), ('update_metadata', 0)]
    >>> kw = dict(idxs='idx1 idx2'.split(), update_metadata=1)
    >>> umri(kw)
    [('idxs', ['idx1', 'idx2']), ('update_metadata', 1)]

    In fact, that `update_metadata` option "rules", if given,
    in affecting the default for the `idxs` option.
    >>> kw = dict(update_metadata=1)
    >>> umri(kw)
    [('idxs', ['getId']), ('update_metadata', 1)]
    >>> kw = dict(update_metadata=0)
    >>> umri(kw)
    [('idxs', None), ('update_metadata', 0)]

    You can explicitly specify update_metadata=None; in this case,
    it depends on the idxs specification:
    >>> kw = dict(update_metadata=None)
    >>> umri(kw)
    [('idxs', None), ('update_metadata', 0)]
    >>> kw = dict(idxs=None, update_metadata=None)
    >>> umri(kw)
    [('idxs', None), ('update_metadata', 0)]
    >>> kw = dict(idxs='idx1 idx2'.split(), update_metadata=None)
    >>> umri(kw)
    [('idxs', ['idx1', 'idx2']), ('update_metadata', 0)]
    >>> kw = dict(idxs=''.split(), update_metadata=None)
    >>> umri(kw)
    [('idxs', ['getId']), ('update_metadata', 1)]
    """
    changes = {}
    # The 'update_metadata' option "rules":
    if 'update_metadata' in kwargs:
        update_metadata = kwargs['update_metadata']
    else:
        update_metadata = 1
        changes.update({
            'update_metadata': update_metadata,
            })

    if 'idxs' in kwargs:
        idxs = kwargs['idxs']
        if update_metadata is None:  # must be specified explicitly
            if idxs is None:  # the known value for "all"
                update_metadata = 0
            elif not idxs:  # "no indexes" is not allowed!
                update_metadata = 1
                idxs = get_default_idxs()
                changes.update({
                    'idxs': idxs,
                    })
            else:
                # we have some "interesting subset" :
                update_metadata = 0
            changes.update({
                'update_metadata': update_metadata,
                })
        if isinstance(idxs, tuple):
            changes.update({
                'idxs': list(idxs),
                })
        elif idxs is None:
            pass
        elif not isinstance(idxs, list):
            changes.update({
                'idxs': idxs.split(),
                })
    elif update_metadata is None:
        # defaults: all indexes, no metadata
        changes.update({
            'idxs': None,
            'update_metadata': 0,
            })
    elif update_metadata:
        changes.update({
            'idxs': get_default_idxs(),
            'update_metadata': 1,
            })
    else:
        changes.update({
            'idxs': None,
            })
    kwargs.update(changes)


def make_reindexer(**kwargs):
    """
    Erzeuge eine Funktion, die das übergebene Objekt reindiziert
    Optionen, alle benannt zu übergeben:

    logger - der zu verwendende Logger
    catalog - der Portalkatalog
    context - der Kontext; benötigt, wenn <catalog> nicht übergeben

    idxs - Liste der zu aktualisierenden Indexe
    update_metadata - sollen die Metadaten aktualisiert werden?
                      (Vorgabe: True)

    Vorgabewerte für die zu erzeugenden Funktion:

    update_metadata - sollen die Metadaten aktualisiert werden?
                      Wenn None, werden die Metadaten aktualisiert,
                      wenn eine leere Liste der Indexe übergeben wird.
    """
    pop = kwargs.pop
    logger = kwargs.pop('logger', None)
    if logger is None:
        logger = logging.getLogger('reindex')
    _update_mri_kwargs(logger, kwargs)
    update_metadata = kwargs.pop('update_metadata')
    ri_kwargs = {'update_metadata': update_metadata,
                 }
    idxs = kwargs.pop('idxs')
    if idxs:
        ri_kwargs['idxs'] = idxs
    catalog = kwargs.pop('catalog', None)
    if catalog is None:
        context = kwargs.pop('context')
        catalog = getToolByName(context, 'portal_catalog')
    catalog_reindex = catalog.reindexObject
    return_brain = kwargs.pop('return_brain', False)
    if return_brain:
        get_brain = catalog._catalog
    _info = [update_metadata and 'update metadata' or 'no metadata',
             idxs and 'indexes: ' + ', '.join(idxs) or 'all indexes',
             return_brain and 'returning brains' or 'returning boolean'
             ]
    debug = kwargs.pop('debug', False)
    if debug:
        _info.append('DEBUG')

    logger.info('make_indexer: %s',
                '; '.join(_info))

    def reindex(brain=None,
                o=None):
        """
        Reindex the given object (by default given as brain!)

        If computing the object yourself, don't give the brain
        but specify the object by name: reindex(o=theobject).
        """
        if o is None:  # Normalfall
            if brain is None:
                raise ValueError('neither brain nor o(bject) given!')
            try:
                o = brain.getObject()
            except AttributeError as e:
                pp(('e:', e), ('brain:', brain), ('o:', o),
                   ('test:', repr(brain).startswith('<PloneSite ')))
                logger.error('KeyError %(e)r for brain (?!) %(brain)r', locals())
                if 0 and 'debug reindex':
                    set_trace()
                    pp(extract_stack())
            except KeyError as e:
                logger.error('KeyError %(e)r for brain %(brain)r', locals())
            except Exception as e:
                logger.error('brain %(brain)r: unexpected exception!',
                             locals())
                logger.exception(e)
            if o is None:
                logger.error("brain %(brain)r: zombie brain?",
                             locals())
                return False
        elif brain is not None:
            o2 = brain.getObject()
            if o2 != o:  # check for identity was too strict!
                logger.error("brain %(brain)r doesn't match object %(o)r",
                             locals())
                return False

        if debug:
            pp('*** Object %(o)r found!' % locals())
            set_trace()

        pt = getattr(o, 'portal_type', None)
        if pt == 'Plone Site':
            logger.warn("%(o)r: Won't reindex the site root", locals())
            return False

        try:
            catalog_reindex(o, **ri_kwargs)
        except POSKeyError as e:
            logger.error('error reindexing %(o)r: %(e)r', locals())
            return False
        except Exception as e:
            logger.error('error reindexing %(o)r: %(e)r', locals())
            raise
        else:
            if return_brain:
                uid = o.UID()
                brains = get_brain(UID=uid)
                if not brains:
                    logger.error('No brains for object %(o)r, uid=%(uid)r!',
                                 locals())
                    return None
                elif brains[1:]:
                    num = len(brains)
                    logger.error('Multiple brains for object %(o)r, uid=%(uid)r (%(num)d)!',
                                 locals())
                    return None
                else:
                    return brains[0]
            return True

    return reindex


def reindex_all(**kwargs):
    """
    Reindiziere die angegebenen Objekte.

    Mögliche benannte Argumente:

    - von make_query_extractor() extrahierte Suchparameter,
      als da wären

      - getCustomSearch
      - getExcludeFromSearch
      - Language
      - path
      - portal_type

    - Argumente zum Erzeugen eines Reindexers

    ...
    """
    logger = kwargs.pop('logger', None)
    if logger is None:
        logger = logging.getLogger('reindex')
    catalog = kwargs.pop('catalog', None)
    if catalog is None:
        context = kwargs.pop('context')
        catalog = getToolByName(context, 'portal_catalog')
    limit = kwargs.pop('limit', None)
    ri_kwargs = {'catalog': catalog,
                 'logger': logger,
                 }
    for name in [
        'idxs',
        'update_metadata',
        ]:
        if name in kwargs:
            ri_kwargs[name] = kwargs.pop(name)

    extract_query = make_query_extractor(context)
    query = extract_query(kwargs)
    portal_types = query.pop('portal_type')
    Language = query.pop('Language')
    if kwargs:
        logger.error('Unused keyword arguments: %(kwargs)s', locals())
    logger.info('reindex_all: query args are:\n  %s',
                '\n  '.join(['%r=%r' % tup
                             for tup in query.items()
                             ]))

    i = 0
    reindex = make_reindexer(**ri_kwargs)
    transaction.begin()
    try:
        for pt in portal_types:
            for la in Language:
                q = {'portal_type': pt,
                     'Language': la,
                     }
                q.update(query)
                ii = 0
                for brain in catalog(q):
                    if not ii:
                        logger.info('portal_type=%(pt)r, Language=%(la)r ...',
                                    locals())
                    ii += 1
                    if reindex(brain):
                        i += 1
                        if not i % 100:
                            logger.info('committing after %(i)r. change', locals())
                            transaction.commit()
                        if limit is not None and i >= limit:
                            return bool(i)
        return bool(i)
    finally:
        if i % 100:
            logger.info('committing remaining changes; total: %(i)r', locals())
            transaction.commit()


if __name__ == '__main__':
    # Standard library:
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*- äöü vim: sw=4 sts=4 et tw=79
# Python compatibility:
from __future__ import absolute_import

from six.moves import map

# Standard library:
from posixpath import normpath
from string import capitalize


def _traversable_path(path):
    """
    Little helper to remove a leading slash (and trailing slashes as well)
    which the .restrictedTraverse method seems to dislike.

    >>> _traversable_path('/e-learning/')
    'e-learning'

    Note: The method will still raise a KeyError if fht object is not yet there!
    """
    res = normpath(path)
    if res.startswith('/'):
        return res[1:]
    return res


def make_title(id):
    """
    A factory for default titles

    >>> make_title('with-dashes_and_underScores')
    'With Dashes And Underscores'
    """
    raw = id.strip()
    if not raw:
        raise ValueError('Nothing to build a title from: %(id)r' % locals())
    chunks = raw.replace('_', ' ').replace('-', ' ').split()
    res = ' '.join(map(capitalize, chunks))
    if not res:
        raise ValueError('Nothing left to build a title from: %(id)r' % locals())
    return res


if __name__ == '__main__':
    # Standard library:
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*- vim: ts=8 sts=4 sw=4 si et tw=79
"""\
Tools für Formulare
"""

# Python compatibility:
from __future__ import absolute_import, print_function

from importlib_metadata import PackageNotFoundError, version
from six import string_types as six_string_types
from six import text_type as six_text_type
from six.moves.urllib.parse import urlencode, urlsplit, urlunsplit

try:
    version('beautifulsoup4')
except PackageNotFoundError:
    HAS_BEAUTIFULSOUP = False
else:
    HAS_BEAUTIFULSOUP = True

try:
    version('visaplan.plone.infohubs')
except PackageNotFoundError:
    HAS_INFOHUBS = False
else:
    HAS_INFOHUBS = True


if HAS_INFOHUBS:
    # visaplan:
    from visaplan.plone.infohubs.hubs2 import context_and_form_tuple
else:
    context_and_form_tuple = None

# Standard library:
from cgi import parse_qsl
from collections import defaultdict
from datetime import date
from string import strip
from time import strptime

__author__ = "Tobias Herp <tobias.herp@visaplan.com>"

if HAS_BEAUTIFULSOUP:
    # 3rd party:
    from bs4 import BeautifulSoup

try:
    # visaplan:
    from visaplan.tools.lands0 import as_new_list

    # Local imports:
    from visaplan.plone.tools.functions import is_uid_shaped
except ImportError:
    if __name__ == '__main__':
        print('Some tests will fail due to import problem')
    else:
        raise
    UIDCHARS_UNICODE = frozenset(u'0123456789abcdef')
    def is_uid_shaped(s, onerror='raise'):
        if isinstance(s, six_text_type):
            pass
        elif isinstance(s, bytes):
            try:
                s = s.decode('ascii')
            except UnicodeDecodeError:
                return False
        elif onerror == 'raise':
            raise ValueError('String expected: %(s)r'
                             % locals())
        else:
            return False
        return len(s) == 32 and UIDCHARS_UNICODE.issuperset(s)
    def as_new_list(val, splitfunc=None):
        if val is None:
            return []
        if not isinstance(val, six_string_types):
            return list(val)
        if splitfunc is None:
            return [s.strip() for s in val.split(',')]
        return splitfunc(val)
# (mit ../../scripts/run-doctests klappt der Export;
# bei direktem Aufruf mit python oder zopepy funktionieren
# wenigstens die Tests, die weder MockRequest noch is_uid_shaped
# benötigen)
try:
    # Local imports:
    from .mock import MockRequest
except (ImportError, ValueError):
    if __name__ != '__main__':
        raise


__all__ = ('tryagain_url',
           'back_to_referer',
           'merge_qvars',
           # -------------------- [ nun in visaplan.tools.lands0 ... [
           # 'list_of_strings',
           # 'string_of_list',
           # 'as_new_list',
           # 'lines_to_list',
           # -------------------- ] ... nun in visaplan.tools.lands0 ]
           # --------------------- [ nun in visaplan.tools.dicts ... [
           # 'subdict',
           # 'subdict_onekey',
           # 'subdict_forquery',
           # --------------------- ] ... nun in visaplan.tools.dicts ]
           'sorted_flaglist',
           'dict_from_form',
           'get_dict',
           'restricted_default',
           'form_default__factory',
           # ----------------- [ nun in visaplan.tools.minifuncs ... [
           # 'gimme_None',
           # 'gimme_True',
           # 'gimme_False',
           # 'gimmeConst__factory',
           # ----------------- ] ... nun in visaplan.tools.minifuncs ]
           # --------------------- [ nun in visaplan.tools.dates ... [
           # 'make_date_parser',
           # 'parse_date',
           # --------------------- ] ... nun in visaplan.tools.dates ]
           'detect_duplicates',
           'uid_or_number',
           # HTML-Code erzeugen:
           'make_input',  # ... but see the new make_input_text
                          # function in visaplan.plone.search.utils!
           # --------------------- [ nun in visaplan.tools.dicts ... [
           # 'updated',
           # --------------------- ] ... nun in visaplan.tools.dicts ]
           )

# hier nicht mehr verwendet, wg. urlsplit;
# wird demnächst entfallen:
QUERY_ONLY_CHARS = frozenset('?&=')


def tryagain_url(request, varnames=None,  # ----- [ tryagain_url ... [
                 path=None,
                 # bitte stets benannt angeben:
                 var_items=None,
                 use_suffix=None,
                 use_first=None):
    """
    Umleitung zur sendenden Seite, um es nochmal zu probieren.

    request -- das Request-Objekt mit form-Dictionary und HTTP_REFERER
    varnames -- Positivliste von Variablen für den Query-String
    path -- Pfad zu einer zu verwendenden Seite/Methode;
            darf keine Fragezeichen o.ä. enthalten
            (als absolute Angabe interpretiert)

    ACHTUNG: Ab visaplan.plone.tools 1.2 müssen alle Argumente außer dem ersten
             (<request>) benannt übergeben werden!
    NOTE: Starting from visaplan.plone.tools 1.2, all arguments except the first
          one (the request) must be given by name!

    >>> req = MockRequest('http://test.me/desktop/tan', tan='123',
    ...                   date='1.1.1970', other='ignoreme')
    >>> tryagain_url(req, ['tan', 'date'])
    'http://test.me/desktop/tan?tan=123&date=1.1.1970'
    >>> tryagain_url(req, ['tan', 'date'], 'schreibtisch/mytan')
    'http://test.me/schreibtisch/mytan?tan=123&date=1.1.1970'

    Von einem übergebenen Pfad <path> wird nur die Pfadkomponente verwendet,
    nicht jedoch ein etwa enthaltener Hostname usw.:

    >>> tryagain_url(req, ['tan', 'date'],
    ...              'http://www.test.me/nur/der/pfad')
    'http://test.me/nur/der/pfad?tan=123&date=1.1.1970'
    >>> tryagain_url(req, ['tan'],
    ...              'manage_xy')
    'http://test.me/manage_xy?tan=123'

    Für Listenwerte ist das Standardverhalten, dem Variablennamen das Suffix
    ':list' anzuhängen und alle nicht-leeren Werte zu verwenden:
    >>> req = MockRequest('http://test.me/desktop/tan',
    ...                   ids=['eins', '', 'zwei'])
    >>> tryagain_url(req, ['ids'])
    'http://test.me/desktop/tan?ids%3Alist=eins&ids%3Alist=zwei'

    Dies kann sich noch ändern; die Sinnhaftigkeit solcher URLs wurde noch
    nicht überprüft!

    Wenn use_first angegeben wird, wird nur der jeweils erste nicht-leere Wert
    weitergereicht, und ':list' wird per default nicht verwendet:
    >>> tryagain_url(req, ['ids'], use_first=True)
    'http://test.me/desktop/tan?ids=eins'

    Alternativ zur Angabe der Variablennamen ist es auch möglich,
    die zu verwendenden Variablen mit ihren Werten zu übergeben:
    >>> tryagain_url(req, var_items=[('name', 'value')])
    'http://test.me/desktop/tan?name=value'
    """
    referer = request['HTTP_REFERER']
    su = list(urlsplit(referer))
    qsl = []
    form = request.form
    if var_items is not None:
        if varnames is not None:
            raise TypeError('Specify *either* varnames *or* var_items!')
        if not isinstance(var_items, list):
            qsl.extend(list(var_items))
        else:
            qsl.extend(var_items)
    if varnames is None:
        varnames = []
    for name in varnames:
        val = form.get(name, None)
        # print 'tryagain_url: %(name)r, %(val)r' % locals()
        if val:
            if isinstance(val, six_string_types):
                qsl.append((name, val.strip()))
            elif isinstance(val, (list, tuple)):
                # done = False
                if use_suffix is None:
                    use_suffix = not use_first
                if use_suffix:
                    name += ':list'
                for chunk in val:
                    if chunk:
                        qsl.append((name, chunk))
                        if use_first:
                            # done = True
                            break
            else:
                qsl.append((name, str(val)))
    su[3] = urlencode(qsl)
    su[4] = ''
    if path:
        # (--> scheme, netloc, path, query, fragment):
        parsed_path = urlsplit(path)
        path = parsed_path.path
        if path:
            su[2] = path
    return urlunsplit(su)
    # ------------------------------------------- ] ... tryagain_url ]


def back_to_referer(context=None,  # --------- [ back_to_referer ... [
                    request=None, url=None,
                    detect_browser=None,
                    redirect=True,
                    items=(),
                    **kwargs):
    """
    Die übliche Umleitung zur verweisenden Seite.

    Standardverwendung:

        return back_to_referer(context)

    am Ende von Browser-Methoden; alle Optionen (außer einem als erstes
    übergebenen Kontext) werden benannt übergeben.

    Modifikation der URL, typischerweise des Referers:
    Es können entweder <items> angegeben werden (eine Sequenz von 2-Tupeln)
    oder benannte Argumente.

    Optionen:
      context - üblicherweise vorhanden und, sofern nicht ohnehin der Request
                verwendet wurde, als unbenanntes Argument übergeben
      request - der eigentlich interessierende Teil des Kontexts.
                Wenn ermittelt, als benanntes Argument übergeben.
      url -     z. B. ein Methodenname.
        ACHTUNG: Es findet keinerlei Behandlung für Query-Strings etc.
                 statt!
      detect_browser -- neu, und daher *vorerst* per Vorgabe nicht aktiv:
                stelle fest, ob der Request '/@@' enthält (--> dann fehlt ein
                korrekter Kontext!) und entfernt ggf. alles ab dieser
                Zeichenkette
      redirect - nur für die Testbarkeit gedacht
      items -   Sequenz von (Variable, Wert)-Tupeln für den Query-String

    Die Option <redirect> dient allein der Testbarkeit:

    >>> req = MockRequest('http://test.me/transmin',
    ...                   actual_url='http://test.me/@@mogrify/import_tarball')
    >>> def b2r_url(context=None, **kwargs):
    ...     kwargs['redirect'] = False
    ...     return back_to_referer(context, **kwargs)

    Ohne detect_browser (aktuelles Standardverhalten) wird die übergebene "url"
    (d.h., meistens nur der Name einer Sicht) direkt verwendet; der Nutzwert
    besteht dann in der Ausführung der Umleitung:

    >>> b2r_url(request=req, url='formname', detect_browser=False)
    'formname'

    Wenn eine Browser-Methode ausgeführt wird, ist das mutmaßlich nicht gut
    genug; die relative Umleitung wird von Zope nicht auf der Basis des Referers
    ausgeführt, sondern der Methode im Kontext. Um eine Umleitung nach
    .../@@browsername/formname zu vermeiden (ohne gültigen Kontext!),
     detect_browser=True angeben:

    >>> b2r_url(request=req, url='formname', detect_browser=True)
    'http://test.me/formname'

    """
    if request is None:
        request = context.REQUEST
    url_given = url is not None
    if url is None:
        url = request['HTTP_REFERER']
        if detect_browser is None:
            detect_browser = False  # Bleibt womöglich
    elif detect_browser is None:
        detect_browser = False  # --- Vorerst - kann sich ändern!
    if detect_browser:
        actual_url = request['ACTUAL_URL']
        if '/@' in actual_url:
            parsed_path = urlsplit(actual_url)
            ppli = list(parsed_path)
            ppp = parsed_path.path
            po = ppp.find('/@@')
            if po == 0:
                if url_given:
                    ppli[2] = '/' + url.lstrip('/')
                else:
                    ppli = list(urlsplit(url))  # der Referer
            elif po != -1:
                if url_given:
                    ppli[2] = ppp[:po+1] + url.lstrip('/')
                else:
                    ppli[2] = ppp[:po]
            url = urlunsplit(ppli)

    if items:
        assert not kwargs
        url = merge_qvars(url, items)
    elif kwargs:
        url = merge_qvars(url, kwargs)
    if not redirect:
        return url
    # print '*** Umleitung -->', url
    return request.RESPONSE.redirect(url)
    # ---------------------------------------- ] ... back_to_referer ]


def merge_qvars(url, items):  # ------------------ [ merge_qvars ... [
    """
    Ergänze eine als String übergebene URL durch die übergebenen
    Query-Variablen.

    >>> merge_qvars('http://test.me/here', (('eins', 1),))
    'http://test.me/here?eins=1'
    >>> merge_qvars('http://test.me/here?eins=0', (('eins', 1),))
    'http://test.me/here?eins=1'
    >>> merge_qvars('http://test.me/here?eins=1', (('zwei', 2),))
    'http://test.me/here?eins=1&zwei=2'

    Es kann auch ein dict-Objekt übergeben werden; in diesem Falle
    werden dessen Schlüssel sortiert:

    >>> merge_qvars('http://test.me/here', {'zwei': 2, 'drei': 3})
    'http://test.me/here?drei=3&zwei=2'

    Es ist möglich, eine Variable zu löschen,
    indem als Wert None übergeben wird:

    >>> merge_qvars('http://test.me/here?eins=0', (('eins', None),))
    'http://test.me/here'

    Die Variablennamen werden in folgender Reihenfolge berücksichtigt:
    - erst die schon in der URL vorhandenen, unter Beibehaltung der
      Reihenfolge;
    - dann die im items-Argument übergebenen (wenn eine Sequenz, unter
      Beibehaltung der gegebenen Reihenfolge; wenn ein dict, in
      ASCII-sortierter Folge)
    Es werden auch URLs ohne Hostkomponente unterstützt.

    >>> merge_qvars('methode?eins=1&zwei=2',
    ...             (('vier', 4), ('drei', 3)))
    'methode?eins=1&zwei=2&vier=4&drei=3'
    >>> merge_qvars('methode?eins=1&zwei=2',
    ...             {'vier': 4, 'drei': 3})
    'methode?eins=1&zwei=2&drei=3&vier=4'

    Mehrfach vergebene Variablennamen werden - nach der Logik der Initiali-
    sierung eines Dictionarys aus einer Sequenz von (key, val)-Tupeln - auf
    den letzten enthaltenen Wert reduziert;
    es gibt keinerlei Listenfunktionalität.

    Wenn Listen erhalten bleiben sollen, ist dafür eine spezielle Funktion
    nötig!

    Hier ein Beispiel zur Illustration, was implizit passiert:

    >>> merge_qvars('http://test.me/here?eins=0&eins=1', ())
    'http://test.me/here?eins=1'

    Da dieses Verhalten durchaus gewünscht sein und dazu genutzt werden kann,
    URLs zu "normalisieren", sollte es nicht leichtfertig geändert werden.
    """
    given_dic = dict(items)
    if isinstance(items, dict):
        items = sorted(items.items())
    given_url = urlsplit(url)
    changable = list(given_url)
    # in der geg. URL vorhandene Query-Variablen
    # (zunächst als Liste, um die Reihenfolge zu erhalten):
    query_lst = parse_qsl(given_url.query)
    query_dic = dict(query_lst)
    new_query = []
    query_dic.update(given_dic)
    # zuerst die vorhandenen, dann die gegebenen Variablen,
    # jeweils in Originalreihenfolge:
    for lst in (query_lst, items):
        for tup in lst:
            key, oldval = tup
            if key not in query_dic:
                continue
            val = query_dic.pop(key)
            if val is not None:
                new_query.append((key, val))
    changable[3] = urlencode(new_query)
    return urlunsplit(changable)
    # -------------------------------------------- ] ... merge_qvars ]


# ------------------------------- [ nun in visaplan.tools.lands0 ... [
# def list_of_strings(): hier gelöscht
# def string_of_list(): hier gelöscht
# def lines_to_list(): hier gelöscht
# def as_new_list(): hier gelöscht
# ------------------------------- ] ... nun in visaplan.tools.lands0 ]


# -------------------------------- [ nun in visaplan.tools.dicts ... [
# def subdict(): hier gelöscht
# def subdict_onekey(): hier gelöscht
# def subdict_forquery(): hier gelöscht
# -------------------------------- ] ... nun in visaplan.tools.dicts ]


def sorted_flaglist(val, keys, first=None, strict=True):
    """
    Z. B. um TTW gewählte Aktionen auszuwerten und in einer Sequenz von
    Checkboxen darzubieten:

    >>> sorted_flaglist('parse', ('parse', 'fix'))
    [('parse', True), ('fix', False)]

    Die in <first> angegebenen Schlüssel sind immer True:
    >>> sorted_flaglist('parse', ('parse', 'fix'), 'bonk')
    [('bonk', True), ('parse', True), ('fix', False)]
    """
    res = []
    val = as_new_list(val)
    first = as_new_list(first)
    done = set()
    for key in first:
        if key in done:
            continue
        res.append((key, True))
        done.add(key)
    for key in keys:
        if key in done:
            continue
        if key in val:
            res.append((key, True))
        else:
            res.append((key, False))
        done.add(key)
    if strict:
        return res
    for key in val:
        if key in keys or key in done:
            continue
        res.append((key, False))
        done.add(key)
    return res


def dict_from_form(form, name, names=None):
    """
    Lies ein dict aus dem Request;
    siehe <http://old.zope.org/Members/Zen/howto/FormVariableTypes>.

    Wir probieren zwei Varianten, ein dict per Formular zu erzeugen:

    - direkt: dann wird der Schlüsselname als Feldname verwendet
      (geht nur, wenn die Schlüsselnamen vorab bekannt sind, also
       serverseitig erzeugt werden);
    - über eine Liste von dicts; in diesem Fall müssen zwei Namen angegeben werden,
      um die Schlüssel und die Werte zuordnen zu können.

    >>> src = {'aslist': [{'key': 'eins', 'val': 'zwei'}],
    ...        'asdict': {'eins': 'zwei', 'drei': None},
    ...        }
    >>> dict_from_form(src, 'aslist', ('key', 'val'))
    {'eins': 'zwei'}
    >>> dict_from_form(src, 'asdict')
    {'eins': 'zwei'}
    """
    val = form.get(name)
    if not val:
        return {}
    if isinstance(val, (list, tuple)):
        # eine Sequenz von dicts:
        keyname, valname = names
        val = dict([(dic[keyname], dic[valname])
                    for dic in val])
    return {k: v
            for (k, v) in val.items()
            if v}


def get_dict(a, fork=False):
    """
    Gib ein übergebenes dict-Objekt zurück
    oder erzeuge es.

    Das Argument ist entweder schon ein dict-Objekt:

    >>> dic_a = {'a': 1}
    >>> dic_b = get_dict(dic_a)
    >>> dic_b is dic_a
    True
    >>> dic_c = get_dict(dic_a, True)
    >>> dic_c is dic_a
    False
    >>> dic_c == dic_a
    True

    ... oder eine Sequenz [f[, args][, kwargs]]
    mit einer Funktion f, die (hoffentlich; hier nicht überprüft!)
    mit f(*args, **kwargs) ein dict-Objekt erzeugt:

    >>> get_dict([dict])
    {}
    >>> get_dict([dict, ((('b', 2),),)])
    {'b': 2}

    >>> def df(a, b):
    ...   return locals()
    >>> get_dict([df, (11,), {'b': 22}])
    {'a': 11, 'b': 22}
    """
    if isinstance(a, dict):
        if fork:
            return dict(a)
        return a

    has_args = False
    has_kwargs = False
    needs_func = True
    f = None
    args = ()
    kwargs = {}
    for arg in a:
        if needs_func:
            f = arg
            needs_func = False
        elif isinstance(arg, tuple):
            if has_args:
                raise ValueError('%(f)r(%(arg)r): ist zuviel'
                                 % locals())
            args = arg
            has_args = True
        elif isinstance(arg, dict):
            if has_kwargs:
                raise ValueError('%(f)r(%(arg)r): ist zuviel'
                                 % locals())
            kwargs = arg
            has_kwargs = True
        else:
            raise ValueError('%(f)r(%(arg)r): Tupel oder dict erwartet'
                             % locals())
    return f(*args, **kwargs)


def restricted_default(form, key, choices, default=None):
    """
    Gib den Wert für den Schlüssel <key> zurück, sofern er in <choices>
    enthalten ist; ansonsten <default>.

    >>> form = {'action': 'view',
    ...         'evil': 'erase_all'}
    >>> choices = ('create', 'edit', 'view')
    >>> restricted_default(form, 'action', choices)
    'view'
    >>> restricted_default(form, 'evil', choices)
    """
    val = form.get(key)
    if val in choices:
        return val
    return default


def form_default__factory(form, key, choices):
    # shallow copy sollte reichen:
    saved = dict(form)
    def given_or_default(default):
        return restricted_default(saved, key, choices, default)
    return given_or_default


# ---------------------------- [ nun in visaplan.tools.minifuncs ... [
# def gimme_None(): hier gelöscht
# def gimme_True(): hier gelöscht
# def gimme_False(): hier gelöscht
# def gimmeConst__factory(): hier gelöscht
# ---------------------------- ] ... nun in visaplan.tools.minifuncs ]


# -------------------------------- [ nun in visaplan.tools.dates ... [
# def make_date_parser(): hier gelöscht
# parse_date = make_date_parser()
# -------------------------------- ] ... nun in visaplan.tools.dates ]

def detect_duplicates(*args, **kwargs):
    """
    Stelle fest, ob ein oder mehrere Werte der Sequenzvariablen <varname>
    mehrfach vorkommen; das kann bei Zope-Formularen zur Konsequenz haben, daß
    andere Werte unerwarteterweise als Listen ankommen.

    Wenn die Variable bereits ausgelesen wurde, kann der einfachste Aufruf
    verwendet werden:
    >>> list(detect_duplicates('a b c a'.split()))
    ['a']

    Ansonsten werden benannte Argumente benötigt, z. B.:

    >>> list(detect_duplicates('ids', form={'ids': 'a a a b b c'.split()}))
    ['a', 'b']

    - form -- das Formularobjekt. Wenn übergeben, wird ein etwaiges unbenanntes
              Argument als <varname> interpretiert, sonst als <values>;
              wenn <values> fehlt, muß <form> einen Schlüssel <varname>
              enthalten
    - context -- der Kontext
    - varname -- der Name der Variablen.
                 <form> wird benötigt und ggf. aus <context> ermittelt.
    - values -- ./. varname

    """
    form = kwargs.pop('form', None)
    context = kwargs.pop('context', None)
    try:
        values = kwargs.pop('values')
    except KeyError:
        try:
            varname = kwargs.pop('varname')
        except KeyError:
            # weder values noch varname explizit angegeben.
            # Wenn <form> explizit angegeben wurde,
            # wird das unbenannte erste Argument als <varname> interpretiert,
            # sonst als <values>
            # (und wenn es nicht existiert, tritt ein Fehler auf):
            if form is not None:
                varname = args[0]
                values = form[varname]
            else:
                values = args[0]
            assert not args[1:]
        else:                 # varname angegeben,
            if form is None:  # form nicht: context muß vorhanden sein
                form = context.REQUEST.form
            values = form[varname]   # ohne <values>
            assert not args
    else:
        assert 'varname' not in kwargs
        assert not args

    if isinstance(values, six_string_types):
        raise ValueError('Variable %(varname)s contains a string'
                         ' (sequence expected): %(values)r'
                         % locals())
    cnt = defaultdict(int)
    for val in values:
        cnt[val] += 1
        if cnt[val] == 2:
            yield val


def uid_or_number(val, **kwargs):
    """
    Gib ein dict zurück mit den Schlüsseln 'uid', 'number' und 'prefix';
    diese Schlüssel werden je nach Eingabewert befüllt:

    >>> uid_or_number('42')['number']
    42
    >>> uid_or_number('12345678123456781234567812345678')['uid']
    '12345678123456781234567812345678'

    Als Eingabewerte werden auch Ganzzahlen akzeptiert:
    >>> uid_or_number(42)['number']
    42

    Die notorischen Präfixe ('uid-', 'rel-') werden erkannt:
    >>> dic = uid_or_number('rel-12345678123456781234567812345678')
    >>> sorted(dic.items())
    [('number', None), ('prefix', 'rel-'), ('uid', '12345678123456781234567812345678')]

    Um die Fehlinterpretation von UIDs als Zahlen zu unterbinden, gibt es ein
    <limit> für konvertierte Zahlen:
    >>> uid_or_number('10001')
    Traceback (most recent call last):
        ...
    ValueError: 10001 exceeds the limit! (10000)

    Das Limit kann explizit angegeben oder auch aufgehoben werden:
    >>> uid_or_number('10001', limit=None)['number']
    10001

    Wenn eine Zahl übergeben wurde, gibt es keine solche Beschränkung:
    >>> uid_or_number(10001)['number']
    10001

    We might use this as well to prefix our lesson ids in course xml:
    >>> sorted(uid_or_number('lesson-5').items())
    [('number', 5), ('prefix', 'lesson-'), ('uid', None)]
    """
    res = {'uid':    None,
           'number': None,
           'prefix': '',
           }
    if val is None:
        return res
    elif isinstance(val, int):
        res['number'] = val
        return res
    liz = val.split('-', 1)
    if liz[1:]:
        res['prefix'] = liz[0] + '-'
        val = liz[1]
    num = None
    try:
        num = int(val)
        limit = kwargs.pop('limit', 10000)
        if limit is not None and num > limit:
            raise ValueError('%(num)r exceeds the limit! (%(limit)r)'
                             % locals())
        res['number'] = num
    except ValueError:
        if is_uid_shaped(val):
            res['uid'] = val
        else:
            raise
    return res


def make_input(data, **kwargs):
    """
    Erzeuge HTML-input-Elemente für die übergebenen Daten <data>.

    Das optionale Schlüsselwortargument <type> gibt den Typ an (Vorgabe:
    "hidden");
    sonstige Schlüsselwortargumente werden an den BeautifulSoup-Konstruktor
    übergeben; hier für die Doctests:
    >>> kw={'features': 'html.parser'}

    Hauptzweck ist es, hidden-Felder zu erzeugen:
    >>> make_input({'getCustomSearch': ['portal_type=UnitraccImage']}, **kw)
    '<input name="getCustomSearch:list" type="hidden" value="portal_type=UnitraccImage"/>'
    >>> make_input({'depth': 3}, **kw)
    '<input name="depth:int" type="hidden" value="3"/>'

    Für leere Listen wird dennoch ein (leerer) Eintrag erzeugt
    (sofern nicht irgendwie ignore_empty angezeigt ist; derartiges ist hier
    aber noch nicht implementiert)

    >>> make_input({'emptylist': []}, **kw)
    '<input name="emptylist:list" type="hidden" value=""/>'

    Der hidden-Typ kann durch explite Angabe einer "falschen" type-Option
    unterdrückt werden:
    >>> make_input({'getCustomSearch': ['portal_type=UnitraccImage']}, type=None, **kw)
    '<input name="getCustomSearch:list" value="portal_type=UnitraccImage"/>'
    """
    # TODO: leere Elemente (Whitelist/Blacklist, :ignore_empty);
    #       vermutlich mit anderer Funktionssignatur oder Funktions-Factory
    if not data:
        return None
    # das klappt nicht: (!)
    # assert isinstance(data, dict), \
    #       'dictionary expected (%r)' % (data,)
    type = kwargs.pop('type', 'hidden')
    if type:
        def attr_dict(name, value):
            return {'name': name,
                    'value': str(value),
                    'type': type,
                    }
    else:
        def attr_dict(name, value):
            return {'name': name,
                    'value': str(value),
                    }
    if kwargs and 0:
        raise TypeError('Currently all kwargs ignored! (%s)'
                        % (kwargs,))
    soup = BeautifulSoup(**kwargs)
    res = []
    for key, val in data.items():
        # Siehe http://old.zope.org/Members/Zen/howto/FormVariableTypes,
        # http://dev-wiki.unitracc.de/wiki/Typenkonversion_in_Zope-Formularen
        if isinstance(val, list):
            key += ':list'
            empty = True
            for thisval in val:
                if thisval not in ('', None):
                    empty = False
                    tag = soup.new_tag('input')
                    tag.attrs = attr_dict(key, thisval)
                    res.append(str(tag))
            if empty:  # XXX checkme; wo ist das dokumentiert?!
                tag = soup.new_tag('input')
                tag.attrs = attr_dict(key, '')
                res.append(str(tag))
        elif isinstance(val, six_string_types):
            tag = soup.new_tag('input')
            tag.attrs = attr_dict(key, val)
            res.append(str(tag))
        elif isinstance(val, int):
            key += ':int'
            tag = soup.new_tag('input')
            tag.attrs = attr_dict(key, val)
            res.append(str(tag))
        else:
            raise ValueError('make_input, key=%(key)r:'
                             ' type of %(val)r unsupported'
                             % locals())
    return '\n'.join(res)


# -------------------------------- [ nun in visaplan.tools.dicts ... [
# def updated(): hier gelöscht
# -------------------------------- ] ... nun in visaplan.tools.dicts ]


def form_changes(hub=None, info=None, context=None, form=None, func=None):
    """
    Return a (form, changes, deletions) tuple, suitable as input for
    visaplan.tools.dicts.update_dict.

    A function to calculate necessary changes to form data;
    the signature is tailored after the first example in the utils module of
    visaplan.plone.industrialsector (see there); functions of this type are
    aimed to establish a plugin structure for search forms.

    Requires visaplan.plone.infohubs.

    Options are:
        hub, info -- as returned by visaplan.plone.infohubs.make_hubs
                     (if you have them already);
                     alternatively, you'll usually specify the context
                     or -- e.g. after calling sibling functions which will
                     return it -- the form.
        context -- the usual context
        form -- the form data dictionary, usually context.REQUEST.form
                (modified in-place).
        func -- a function (not used in this stub).
                Might become removed in future versions, or perhaps replaced by
                general support for *args and/or **kwargs.

    This is a stub; `changes` will always be an empty dict,
    `deletions` will always be an empty list:

    >>> req = MockRequest(text='seekme')
    >>> form_changes(form=req.form)
    ({'text': 'seekme'}, {}, [])
    """
    if context_and_form_tuple is not None:
        hub, info, context, form = context_and_form_tuple(hub, info, context, form,
                                                          amend=True)
    else:  # for better testability
        assert form is not None
    changes = {}
    deletions = []
    return (form, changes, deletions)


if __name__ == '__main__':
    class MockRequest(dict):
        # kopiert ins mock-Modul
        def __init__(self, referer=None, **kwargs):
            self['HTTP_REFERER'] = referer
            self['ACTUAL_URL'] = kwargs.pop('actual_url', referer)
            self.form = kwargs

    # Standard library:
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*- äöü vim: sw=4 sts=4 si et textwidth=72 cc=+8
"""
visaplan.plone.tools.log - Helferlein für Entwicklung und Debugging

Autor: Tobias Herp
"""

# Python compatibility:
from __future__ import absolute_import

from six import string_types as six_string_types
from six.moves import map

# Standard library:
from contextlib import contextmanager
from fcntl import LOCK_EX, LOCK_UN, lockf
from functools import wraps
from os import getpid, linesep
from pprint import pformat
from time import strftime

try:
    # Zope:
    from Globals import DevelopmentMode
except ImportError:
    # Hotfix for Zope 4; how to properly replace this?
    DevelopmentMode = False

# visaplan:
from visaplan.tools.coding import safe_decode, safe_encode

# Local imports:
from visaplan.plone.tools.cfg import get_debug_active, split_filename

# Logging / Debugging:
import logging
from visaplan.tools.debug import arginfo, pretty_funcname

__all__ = ['getLogSupport',
           'make_textlogger',
           'locked_open',  # a context manager
           ]

TIMEFORMAT_VERBOSE = u'%Y-%m-%d %T %Z (%A) '
TIMEFORMAT_LASTCHARS = u' |-\t'
_DEVMODE_STRING = str(DevelopmentMode)

def getLogSupport(name=None,
                  fn=None,
                  defaultFromDevMode=False,
                  default=None,
                  **kwargs):
    """
    Verwendung:

      from ...tools.log import getLogSupport
      logger, debug_active, DEBUG = getLogSupport(fn=__file__)
    oder
      logger, debug_active, DEBUG = getLogSupport()

    Standardmäßig Debugging aktivieren:
      logger, debug_active, DEBUG = getLogSupport(..., default='True')

    ... oder nur bei Entwicklungsmodus:
      logger, debug_active, DEBUG = getLogSupport(..., defaultFromDevMode=1)
    """
    if 'stack_limit' not in kwargs:
        kwargs['stack_limit'] = 3
    name, label = split_filename(fn, name,  # stack_limit=stack_limit,
                                 **kwargs)
    logger = logging.getLogger(label)
    if default is None:
        default = defaultFromDevMode and _DEVMODE_STRING or 'False'
    elif not isinstance(default, six_string_types):
        raise ValueError('default must be a string! (%(default)r)'
                         % locals())
    tmp = get_debug_active(name,
                           default)
    if DevelopmentMode:
        debug_active = tmp
    else:
        # Debugging auch in Produktionsmodus:
        # Wert auf 2 (oder noch höher) setzen
        debug_active = tmp > 1
    if debug_active:
        method = logger.info
        method('*** Debugging AKTIV ***')
    else:
        method = logger.debug
    return (logger,
            debug_active,
            method,
            )  # ----------------------------------- ... getLogSupport

@contextmanager
def locked_open(filename, mode='r'):
    """
    Öffnet die übergebene Datei und reserviert sie für exklusiven Zugriff;
    zu verwenden wie die eingebaute open-Funktion
    """
    # von --> https://gist.github.com/lonetwin/7b4ccc93241958ff6967
    with open(filename, mode) as fd:
        lockf(fd, LOCK_EX)
        yield fd
        lockf(fd, LOCK_UN)


def make_textlogger(fn, sorting_filter=None, divide=None, lock=None,
                    time_format=TIMEFORMAT_VERBOSE,
                    debug=False):
    """
    Erzeuge eine Funktion, die (mit <lock>) eine Datei zum exklusiven Zugriff
    reserviert, einmal schreibt und dann nach Freigabe der Datei wieder
    zurückkehrt.

    fn -- der Dateiname, z. B. '.../users.txt'
    sorting_filter -- Eine Funktion, die die Schlüssel eines übergebenen
                      Python-Dicts sortiert und ggf. filtert.
    divide -- eine Trennzeile
    lock -- soll die Datei jeweils zum exklusiven Zugriff gesperrt
            werden? (None --> Default verwenden)
    time_format -- Vorgabe: '%Y-%m-%d %T %Z (%A) '
    debug -- wenn True, werden die internen Aufrufe von make_text
             protokolliert

    Es wird jedenfalls ein Logger erzeugt, der über den Vorgang
    informiert und (für den Fall, daß die Dateiangabe leer ist) als
    Fallback-Option fungiert.
    """
    logger = getLogSupport(stack_limit=3)[0]
    if sorting_filter is None:
        sorting_filter = dict.items
    if divide is None:
        divide = 79 * u'-'
    elif not divide:
        divide = None
    else:
        divide = safe_decode(divide.strip())
    if not time_format:
        time_format = TIMEFORMAT_VERBOSE
    else:
        time_format = safe_decode(time_format)
    if not time_format[-1] in TIMEFORMAT_LASTCHARS:
        timeformat += TIMEFORMAT_LASTCHARS[0]
    PROTOMASK = u'%%-%ds%%s'

    def fallback(txt, **kwargs):
        """
        Verwende die allgemeine Logging-Unterstützung
        """
        items = sorting_filter(kwargs)
        logger.info(' '.join((txt,
                              '; '.join(['%s=%s' % tup
                                         for tup in items])
                              )))

    def make_text(txt, **kwargs):
        """
        Erzeuge einen mehrzeiligen Text (UTF-8-codiert).
        Eingabeargumente werden mit safe_decode zu Unicode decodiert.
        """
        items = sorting_filter(kwargs)
        maxl = max(list(map(len, [t[0] for t in items]))) + 2
        mask = PROTOMASK % maxl
        liz = [u'']
        if divide is not None:
            liz.append(divide)
        liz.append( strftime(time_format)+safe_decode(txt)+u':')
        liz_u = []
        for key, val in items:
            if isinstance(val, six_string_types):
                liz_u.append((safe_decode(key)+u':', safe_decode(val)))
            else:
                liz_u.append((safe_decode(key)+u':', val))
        liz.extend([mask % tup
                    for tup in liz_u
                    ])
        liz.append(u'')
        return safe_encode(u'\n'.join(liz))

    def log_call_and_result(func):
        DEBUG = logger.info
        _funcname = pretty_funcname(func)
        @wraps(func)
        def inner(*args, **kwargs):
            DEBUG('%s(%s) ...', _funcname, arginfo(*args, **kwargs))
            _res = func(*args, **kwargs)
            if not _res:
                DEBUG('%s(...) --> %r', _funcname, _res)
            elif isinstance(_res, tuple):
                DEBUG('%s(...) --> %s values:', _funcname, len(_res))
                for val in _res:
                    DEBUG(pformat(val))
            else:
                DEBUG('%s(...) -->\n%s',
                      _funcname,
                      pformat(_res))
            return _res

        inner.__name__ = _funcname
        return inner

    if debug:
        make_text = log_call_and_result(make_text)

    if lock:
        file_context = locked_open
    else:
        file_context = open

    def text_logger(txt, **kwargs):
        cooked = make_text(txt, **kwargs)
        with file_context(fn, 'a') as fd:
            fd.write(cooked)

    if fn:
        # OK, Dateiname angegeben; "locking logger" erzeugen:
        if lock:
            logger.info('Creating locking logger for %(fn)r', locals())
        else:
            logger.info('Creating text logger for %(fn)r', locals())
        return text_logger

    logger.info('Creating default logger!')
    return fallback


if __name__ == '__main__':
    # Standard library:
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*- äöü vim: ts=8 sts=4 sw=4 si et hls tw=79
# Python compatibility:
from __future__ import absolute_import

from six import text_type as six_text_type

# Zope:
from Products.CMFCore.utils import getToolByName

# visaplan:
from visaplan.tools.lands0 import list_of_strings

# Local imports:
from ._helpers import build_groups_set

try:
    # visaplan:
    from visaplan.tools.coding import safe_decode
except ImportError:
    if __name__ != '__main__':
        raise
    def safe_decode(s):
        if isinstance(s, six_text_type):
            return s
        return s.decode('utf-8')

__all__ = [
    'is_direct_member__factory',
    'is_member_of__factory',
    'is_member_of_any',
    'get_all_members',
    'recursive_members',  # helper for get_all_members
    ]


def is_direct_member__factory(context, userid):
    """
    Erzeuge eine Funktion, die prüft, ob der *beim Erzeugen angegebene* Nutzer
    in der bei jedem Aufruf anzugebenden Gruppe direkt enthalten ist.
    """
    acl = getToolByName(context, 'acl_users')
    gpm = acl.source_groups._group_principal_map

    def is_direct_member_of(group_id):
        return userid in gpm[group_id]

    return is_direct_member_of


def is_member_of__factory(context, userid):
    """
    Erzeuge eine Funktion, die die *direkte oder indirekte* Mitgliedschaft
    des übergebenen Users in der jeweils zu übergebenden Gruppe überprüft.
    """
    acl = getToolByName(context, 'acl_users')
    gpm = acl.source_groups._group_principal_map

    groups = build_groups_set(gpm, userid)

    def is_member_of(groupid):
        return groupid in groups

    return is_member_of


def is_member_of_any(context, group_ids, user_id=None, default=False):
    """
    Is the given or active user member of one of the given groups?

    - For anonymous execution, raises Unauthorized
      (might become changable by keyword-only argument)
    - The normal usage is without specification of the user_id,
      i.e. checking for the logged-in user
    - if the group_ids sequence is empty, the default is used

    """
    pm = getToolByName(context, 'portal_membership')
    if pm.isAnonymousUser():
        raise Unauthorized
    if user_id is not None:
        # wer darf sowas fragen? Manager, Gruppenmanager, ...?
        # TODO: Hier entsprechende Überprüfung!
        pass

    if not group_ids:
        return default
    if user_id is None:
        member = pm.getAuthenticatedMember()
        user_id = member.getId()

    return user_id in get_all_members(context, group_ids)


def get_all_members(context, group_ids, **kwargs):  # --- [[
    """
    Return all members of the given groups

    Liefere alle Mitglieder der übergebenen Gruppe(n).

    Schlüsselwortargumente für .utils.recursive_members und
    groupinfo_factory dürfen angegeben werden;
    letztere werden aber ignoriert, wenn die vorgabegemäße Filterung
    groups_only=True übersteuert wird. In diesem Fall (potentiell sowohl
    Benutzer als auch Gruppen, oder nur Benutzer) werden nur IDs
    zurückgegeben.

    Rückgabe:
    - Sequenz von Gruppeninformationen, mit groups_only=True (Vorgabe);
    - ansonsten nur eine Sequenz der IDs (je nach Aufrufargumenten nur
      Benutzer-IDs, oder gemischt)
    """
    acl = getToolByName(context, 'acl_users')
    gpm = acl.source_groups._group_principal_map
    filter_args = {}
    for key in ('groups_only', 'users_only',
                'containers',
                'default_to_all'):
        try:
            filter_args[key] = kwargs.pop(key)
        except KeyError:
            pass
    members = recursive_members(list_of_strings(group_ids),
                                gpm, **filter_args)
    groups_only = filter_args.get('groups_only', False)
    if groups_only:
        format_args = {'pretty': False,
                       'forlist': True,
                       'missing': 1,
                       }
        format_args.update(kwargs)
        ggibi = groupinfo_factory(context, **format_args)
        res = []
        for gid in members:
            res.append(ggibi(gid))
        return res
    elif kwargs and debug_active:
        pp('ignoriere:', kwargs)

    return members
# ---------------------------------- ] ... get_all_members ]


def recursive_members(gids, dic,
                      containers=None,
                      groups_only=False,
                      users_only=False,
                      default_to_all=False):
    """
    Recursively find all members of the (by id) given groups.

    Mandatory arguments:

      gids -- group ids; a sequence or (for convenience) a string
              (which would be split by whitespace)
      dic -- a dictionary; usually acl.source_groups._group_principal_map

    Ermittle die rekursiv aufgelösten Mitglieder der übergebenen
    Gruppen.

    >>> dic = {'group_a': ['group_b', 'group_c'],
    ...        'group_b': ['user_a', 'user_b'],
    ...        'group_c': ['user_c'],
    ...        'group_d': ['group_a'],
    ...        }
    >>> am = recursive_members(['group_a'], dic, groups_only=True)
    >>> sorted(am)
    ['group_b', 'group_c']
    >>> sorted(recursive_members(['group_a'], dic))
    ['group_b', 'group_c', 'user_a', 'user_b', 'user_c']
    >>> sorted(recursive_members(['group_d'], dic, users_only=True))
    ['user_a', 'user_b', 'user_c']
    >>> sorted(recursive_members(['group_d'], dic, groups_only=True))
    ['group_a', 'group_b', 'group_c']

    containers-Argument:
    - True: die Container werden hinzugefügt
      (mit groups_only: ... sofern es wirklich Gruppen sind)
    - False: die Container werden ausgefiltert
    - None (Vorgabe): weder aktiv hinzufügen noch ausfiltern

    >>> kwargs = {'groups_only': True,
    ...           'containers': True}
    >>> sorted(recursive_members(['group_d'], dic, **kwargs))
    ['group_a', 'group_b', 'group_c', 'group_d']
    >>> kwargs = {'groups_only': True,
    ...           'containers': False}
    >>> sorted(recursive_members(['group_c', 'group_d'], dic, **kwargs))
    ['group_a', 'group_b']
    >>> kwargs = {'users_only': True,
    ...           'containers': True}
    >>> sorted(recursive_members(['group_d'], dic, **kwargs))
    ['user_a', 'user_b', 'user_c']
    >>> kwargs = {'containers': False}

    Das spezielle Argument default_to_all erfordert groups_only=True
    und gibt alle Gruppen zurück, sofern keine Gruppen-IDs übergeben wurden:

    >>> sorted(recursive_members([], dic, groups_only=True,
    ...                          default_to_all=True))
    ['group_a', 'group_b', 'group_c', 'group_d']
    """
    if users_only and groups_only:
        raise ValueError('recursive_members(%(gids)r): '
                         '*either* groups_only '
                         '*or* users_only!'
                         % locals())
    if not gids:
        if default_to_all:
            assert groups_only, 'default_to_all erfordert groups_only!'
            return set(dic.keys())
        else:
            # ohne default_to_all: keine Gruppen, keine Mitglieder
            return set()
    filtered = users_only or groups_only
    res = set()
    exclude = set()
    if containers is None:
        pass
    elif not containers:
        exclude.update(gids)
    for gid in gids:
        try:
            newly_found = set(dic[gid]).difference(res)
            if containers and not users_only:
                res.add(gid)
        except KeyError:  # keine Gruppe, oder?!
            if containers and not groups_only:
                res.add(gid)
        else:
            while newly_found:
                res.update(newly_found)
                this_iteration = set()
                for mid in newly_found:  # member id
                    try:
                        found_here = set(dic[mid]).difference(res)
                        if found_here:
                            this_iteration.update(found_here)
                        if users_only:   # Gruppen aus Ergebnis entfernen
                            exclude.add(mid)
                    except KeyError:
                        if groups_only:  # Benutzer aus Ergebnis entfernen
                            exclude.add(mid)
                res.update(this_iteration)
                newly_found = this_iteration
    res.difference_update(exclude)
    return res




# -*- coding: utf-8 -*- äöü vim: sw=4 sts=4 et tw=79
"""
Tools für Produkt-Setup (Migrationsschritte, "upgrade steps"): _roles
"""

# Python compatibility:
from __future__ import absolute_import

from six import string_types as six_string_types

# visaplan:
from visaplan.tools.classes import Proxy

# Local imports:
from visaplan.plone.tools.setup._args import extract_object_and_brain

__all__ = [
        'make_simple_localroles_function',
        'set_local_roles',
        ]


def make_simple_localroles_function(**kwargs):
    """
    Erzeuge eine einfache Funktion, um (hier: ohne Differenzierung nach
    portal_type ö.ä.) für alle Objekte mit dem angegeben Workflow-Zielstatus
    <target_state> die lokalen Rollen <roles> für den Prinzipal <userid>
    zuzuweisen; die Funktion set_local_roles wird dies dann so interpretieren,
    daß etwaige schon vorhandene weitere lokale Rollen erhalten bleiben.
    """
    for_target_state = kwargs.pop('target_state', 'restricted')
    userid = kwargs.pop('userid')
    roles = kwargs.pop('roles')
    add = kwargs.pop('add', True)
    if isinstance(roles, six_string_types):
        roles = [roles]
    def new_local_roles(brain, target_state):
        if target_state == for_target_state:
            return [(userid, roles, add)]
        return []
    return new_local_roles


def set_local_roles(**kwargs):
    """
    Modifiziere die lokalen Rollenzuweisungen des übergebenen Objekts

    Argumente, alle benannt zu übergeben:

    o -- das Objekt
    brain -- dessen Indexobjekt
             (eines der beiden muß angegeben werden)

    logger -- der Logger

    func -- Eine Funktion, die aus <brain> und <target_state> eine Liste mit
            den gewünschten Änderungen ermittelt (wie <thelist>);
            siehe make_simple_localroles_function
    target_state -- nur zur Übergabe an <func>

    thelist -- eine Liste von (userid, roles [, add])-Tupeln;
               nur benötigt (und verwendet), wenn <func> nicht angegeben

    Von diesen wird zwingend benötigt:
    - logger
    - mindestens eines von o und brain
      (keine explizite Prüfung; es wird aber ggf. ein Fehler auftreten)
    - alternativ:
      - func und target_state, um die zu setzenden oder zu löschenden
        Rollen mit Hilfe der Funktion <func> zu ermitteln;
        oder:
      - thelist
        (derzeit ignoriert, wenn <func> übergeben wurde und nicht None ist)
    """
    o, brain = extract_object_and_brain(kwargs)
    # if 'func' in kwargs:
    func = kwargs.pop('func', None)
    if func is not None:
        # wird benötigt, als Argument für die Funktion!
        target_state = kwargs.pop('target_state')
        thelist = func(brain, target_state)
    else:
        thelist = kwargs.pop('thelist')
    logger = kwargs.pop('logger')
    if not thelist:
        return False
    uid = brain.UID
    changes = 0
    def set_of_roles(userid):
        return set(o.get_local_roles_for_userid(userid=userid))

    roles_of_user = Proxy(set_of_roles)
    changed_users = set()
    for tup in thelist:
        userid, roles = tup[:2]
        if isinstance(roles, six_string_types):
            roles = [roles]
        if tup[2:]:
            tail = tup[3:]
            if tail:
                logger.warn('set_local_roles: %(tup)r is too long'
                            '; ignoring %(tail)r',
                            locals())
            add = tup[2]
            if add not in (True, False,
                           1, 0,  # pep 20.2
                           ):
                txt = ('set_local_roles: %(tup)r contains wrong 3rd value'
                       ' %(add)r; boolean expected'
                       ) % locals()
                logger.error(txt)
                raise ValueError(txt)
        else:
            add = True

        for role in roles:
            roles_set = roles_of_user[userid]
            if role in roles_set:
                if add:
                    logger.info('%(uid)r local roles for %(userid)r:'
                                ' %(role)r already found (%(o)r)', locals())
                else:
                    logger.info('%(uid)r local roles for %(userid)r:'
                                ' removing %(role)r (%(o)r)', locals())
                    roles_set.discard(role)
                    changed_users.add(userid)
            else:
                if add:
                    logger.info('%(uid)r local roles for %(userid)r:'
                                ' adding %(role)r (%(o)r)', locals())
                    changed_users.add(userid)
                    roles_set.add(role)
                else:
                    logger.info('%(uid)r local roles for %(userid)r:'
                                ' %(role)r not found (%(o)r)', locals())

    if not changed_users:
        return False
    done_users = set()
    for userid in sorted(changed_users):
        roles_set = roles_of_user[userid]
        if roles_set:
            sorted_roles = sorted(roles_set)
            logger.info('%(uid)r local roles for %(userid)r:'
                        ' set to %(sorted_roles)s (%(o)r)', locals())
            o.manage_setLocalRoles(userid, sorted_roles)
            done_users.add(userid)
    changed_users.difference_update(done_users)
    if changed_users:
        userids = sorted(changed_users)
        logger.info('%(uid)r local roles for %(userids)s:'
                    ' removing all roles (%(o)r)', locals())
        o.manage_delLocalRoles(userids)
    o.reindexObjectSecurity()
    return True




# -*- coding: utf-8 -*- äöü vim: ts=8 sts=4 sw=4 si et hls tw=79

# Python compatibility:
from __future__ import absolute_import


class PrettyMask(dict):
    """
    For every given "role", suggest a mask for a "pretty" group title.

    >>> PM = PrettyMask(u'%s group "{group}"')
    >>> PM['Reader']
    u'Reader group "{group}"'
    """
    def __init__(self, mask, func=None):
        self._mask = mask
        self._checkfunc = func

    def __getitem__(self, key):
        try:
            return dict.__getitem__(self, key)
        except KeyError:
            val = self._mask % (key,)
            dict.__setitem__(self, key, val)
            return dict.__getitem__(self, key)


PRETTY_MASK = PrettyMask(u'%s group "{group}"')


if __name__ == '__main__':
    # Standard library:
    from doctest import testmod
    testmod()



# -*- coding: utf-8 -*- äöü vim: sw=4 sts=4 et tw=79
"""
Tools für Produkt-Setup (Migrationsschritte, "upgrade steps"): _attr
"""

# Python compatibility:
from __future__ import absolute_import

# visaplan:
from visaplan.tools.classes import CheckedSetterDict, GetterDict, Proxy

__all__ = [
        'make_attribute_setter',
        ]


def make_attribute_setter(logger, setters=None, dryrun=0):
    """
    Gib eine Funktion zurück, die Attribute mit der jeweils passenden Methode
    setzt.

    ACHTUNG: Es wird bisher keinerlei Rückgriff auf das Schema genommen;
             für HTML-Felder z. B. ist das mutmaßlich noch nicht gut genug!
    """
    setters_map = CheckedSetterDict()
    getters_map = GetterDict()
    if setters is not None:
        # spezielle zu verwendende Werte:
        setters_map.update(setters)
    stem = '%(o)r: '

    def make_info(x):
        if x is None:
            return ' using setattr'
        return ' using setter %(x)r' % locals()

    info_dict = Proxy(make_info)

    def set_attribute(o, key, newvals, oldvals=None, dryrun=dryrun):
        """
        Setze das übergebene Attribut und protokolliere den alten und neuen Wert
        """
        val = newvals[key]
        setter_name = setters_map[key]
        setter_info = info_dict[setter_name]
        stem = '%(o)r: '  # sonst nicht gefunden?! §:-|
        if dryrun:
            stem = '<DRYRUN> '+ stem
        msg = stem + 'setting %(key)r to %(val)r%(setter_info)s'
        if oldvals is None:
            getter_name = getters_map[key]
            if getter_name is not None:
                ga = getattr(o, getter_name)
                old = ga()
                msg += ' (was %(old)r)'
        else:
            try:
                old = oldvals[key]
                msg += ' (was %(old)r)'
            except KeyError:
                pass
        if val not in (None, '') or 'old' not in locals():
            logger.info(msg, locals())
        else:
            logger.info(stem + 'deleting %(key)r%(setter_info)s, was %(old)r', locals())
        # dryrun heißt nur: keine Zuweisung; aber ob der Setter existiert,
        # wollen wir durchaus wissen!
        if setter_name is not None:
            a = getattr(o, setter_name)
        if dryrun:
            return
        if setter_name is None:
            setattr(o, key, val)
        else:
            a(val)

    return set_attribute




# -*- coding: utf-8 -*- vim: ts=8 sts=4 sw=4 si et tw=79
"""
visaplan.plone.tools._at1 -- imported via .attools

To replace the use of this one-liner (and use the imported functionality directly),
you can use the following vim script:

%s,^\(\s*\(#\s\?\)\)\(notify\)\(edit\)(\([^) ]\+\))\s*$,from zope import event\rfrom Products.Archetypes.event import ObjectEditedEvent\r\1event.\3(ObjectEditedEvent(\5)),e

This will give you two import lines before each converted code line; you'll need to move those up (and delete duplicates).
"""

# Python compatibility:
from __future__ import absolute_import

# Local imports:
from visaplan.plone.tools._have import HAS_ARCHETYPES

if HAS_ARCHETYPES:
    # Zope:
    from Products.Archetypes.event import ObjectEditedEvent
    from zope import event


def notifyedit(context):
    """
    trigger on-edited events for context
    """
    event.notify(ObjectEditedEvent(context))



# -*- coding: utf-8 -*- äöü vim: ts=8 sts=4 sw=4 si et hls tw=79
# Python compatibility:
from __future__ import absolute_import


def _parse_init_options(kwdict, args):
    """
    Parse the options given during initialization

    We used to have 3 optional arguments which were normally specified
    positionally.  That's not ideal; optional arguments should better be given
    by name.  We don't enforce this yet, for compatibility reasons; but we like
    to have some testability.

    This function takes a dict of keyword arguments, and the tuple of
    positional arguments; it puts all information in that dict, which is
    changed in-place.

    >>> kw = {}
    >>> ar = ()
    >>> _parse_init_options(kw, ar)
    >>> sorted(kw.items())                     # doctest: +NORMALIZE_WHITESPACE
    [('forlist',      0),
     ('missing',      0),
     ('pretty',       0),
     ('title_or_id',  0)]

    The values are just used as they are; we can use this to demonstrate the
    order of the traditional positional options:

    >>> kw.clear()
    >>> _parse_init_options(kw, (1, 2, 3))
    >>> sorted(kw.items())                     # doctest: +NORMALIZE_WHITESPACE
    [('forlist',      2),
     ('missing',      0),
     ('pretty',       1),
     ('title_or_id',  3)]

    If missing=True, we'll need a missing_user_mask as well;
    we have a default value:

    >>> kw={'missing': 1}
    >>> _parse_init_options(kw, ar)
    >>> sorted(kw.items())                     # doctest: +NORMALIZE_WHITESPACE
    [('forlist',      0),
     ('missing',      1),
     ('missing_user_mask', 'Unknown or deleted user "{id}"'),
     ('pretty',       0),
     ('title_or_id',  0)]

    Unknown keys yield TypeErrors, of course:
    >>> kw={'sillykey': 1}
    >>> _parse_init_options(kw, ar)
    Traceback (most recent call last):
      ...
    TypeError: Unsupported keyword option(s) ['sillykey']!

    """
    unsupported = set(kwdict) - set([
        'forlist',
        'missing',
        'pretty',
        'title_or_id'])
    if unsupported:
        unsupported = sorted(unsupported)
        raise TypeError('Unsupported keyword option(s) %(unsupported)s!'
                        % locals())
    if isinstance(args, tuple):
        args = list(args)
    # these used to be accepted positionally:
    if 'pretty' not in kwdict:
        kwdict['pretty'] = (args.pop(0) if args
                            else 0)
    if 'forlist' not in kwdict:
        kwdict['forlist'] = (args.pop(0) if args
                             else 0)
    if 'title_or_id' not in kwdict:
        kwdict['title_or_id'] = (args.pop(0) if args
                                 else 0)
    if args:
        if args[1:]:
            raise TypeError('Unsupported positional options %r (...)'
                            % tuple(args[:1]))
        else:
            raise TypeError('Unsupported positional option %(args)s'
                            % locals())
    # these are new:
    missing_mask_default = 'Unknown or deleted user "{id}"'
    missing_user_mask = kwdict.get('missing_user_mask')
    if 'missing' not in kwdict:
        missing = kwdict['missing'] = int(bool(missing_user_mask))
    else:
        missing = kwdict['missing']
    if missing and not missing_user_mask:
        kwdict['missing_user_mask'] = missing_mask_default


if __name__ == '__main__':
    # Standard library:
    import doctest
    doctest.testmod()



"""
SEO helpers;
for now, supporting structured data, see:

https://developers.google.com/search/docs/appearance/structured-data/intro-structured-data
https://schema.org
"""

# see as well (gf):
# ../../../../../../visaplan.plone.animations/src/visaplan/plone/animations/utils/_lapa.py
# ../../../../../../visaplan.plone.elearning/src/visaplan/plone/elearning/utils/_lapa.py

# Python compatibility:
from __future__ import absolute_import

# Standard library:
from re import match

__all__ = [
    'parse_title',  # ... course, lesson titles
    ]

_LD_CONTEXT = 'https://schema.org'
# -------------------------------- [ regular expression snippets ... [
_RE_START_ANCHOR = (
   u'(?i'      # ignore case (like flags=re.IGNORECASE)
     u'm)'     # multi-line  (like    ...re.MULTILINE)
   u'\\A')     # beginning of the string

_RE_CODE_GROUP = (
    u'(?P<courseCode>'
    u'\\b[A-Z]+'
    u'-'
    u'[0-9]+)'
    u'\\b'
    u' *')
_RE_NAME = (
    u'(?P<name>'
    u'\\b[A-Z].*[A-Za-z)]'
    u')'
    )
_RE_VERSION = (
    u' *- *v[0-9]+\\b'
    )
_RE_FREE = (
    u' *- *[Ff][Rr][Ee][EeIi][!]?'
    )

_RE_END_ANCHOR = u'\\Z'
_RE_END = u' *' + _RE_END_ANCHOR
# -------------------------------- ] ... regular expression snippets ]

RE_LIST = [
    # ----------------------------- [ titles with "course codes" ... [
    # (which we have in lesson titles as well!)
    # title with a version, and optionally "FREE!"
    _RE_START_ANCHOR +
    _RE_CODE_GROUP +
    _RE_NAME +
    _RE_VERSION +
    u'(?:' + _RE_FREE + u')?' +
    _RE_END,
    # title with "FREE!" and perhaps a version:
    _RE_START_ANCHOR +
    _RE_CODE_GROUP +
    _RE_NAME +
    _RE_FREE +
    u'(?:' + _RE_VERSION + u')?' +
    _RE_END,
    # title without known suffixes:
    _RE_START_ANCHOR +
    _RE_CODE_GROUP +
    _RE_NAME +
    _RE_END,
    # ----------------------------- ] ... titles with "course codes" ]

    # ------------- [ the same for titles *WITHOUT* course codes ... [
    # title with a version, and optionally "FREE!"
    _RE_START_ANCHOR +
    _RE_NAME +
    _RE_VERSION +
    u'(?:' + _RE_FREE + u')?' +
    _RE_END,
    # title with "FREE!" and perhaps a version:
    _RE_START_ANCHOR +
    _RE_NAME +
    _RE_FREE +
    u'(?:' + _RE_VERSION + u')?' +
    _RE_END,
    # title without known suffixes:
    _RE_START_ANCHOR +
    _RE_NAME +
    _RE_END,
    # ------------- ] ... the same for titles *WITHOUT* course codes ]
    ]


def parse_title(title, stype=None):
    """
    Parse a given (course, lesson ...) title and return a dict;
    the group names resemble properties for structured types as of schema.org.

    We'll use a little test helper:
    >>> def pt(*args, **kw):
    ...     return sorted(parse_title(*args, **kw).items())

    Our course objects have often a name prefix which resembles a courseCode:

    >>> pt('BE-50 Trench Protection and Pipe Installation - v01')
    ...                # doctest: +NORMALIZE_WHITESPACE
    [(u'courseCode', 'BE-50'),
     (u'name',       'Trench Protection and Pipe Installation')]

    We only return data which we consider useful for structured data use;
    the "- v01" suffix is silently dropped.
    This happens implicitly, since we don't but it in a match group.

    We can use this function to initialize structured data for the respective
    object.
    Let's specify a type:

    >>> pt('BE-50 Trench Protection and Pipe Installation - v01',
    ...    'Course')                   # doctest: +NORMALIZE_WHITESPACE
    [('@context',    'https://schema.org'),
     ('@type',       'Course'),
     (u'courseCode', 'BE-50'),
     (u'name',       'Trench Protection and Pipe Installation')]

    What if we have a lesson here, rather than a course?
    The courseCode wouldn't be valid, so we drop it:

    >>> pt('BE-50 Trench Protection and Pipe Installation - v01',
    ...    'LearningResource')         # doctest: +NORMALIZE_WHITESPACE
    [('@context',    'https://schema.org'),
     ('@type',       'LearningResource'),
     (u'name',       'Trench Protection and Pipe Installation')]

    What if we specify some really unexpected crap?
    At the very least, we return a 'name', with surrounding whitespace
    stripped:

    >>> pt(' (oh such crap!) ')
    [('name', '(oh such crap!)')]

    """
    drop_keys = set()
    if stype:
        res = {
            '@context': _LD_CONTEXT,
            '@type':    stype,
            }
        if stype != 'Course':
            drop_keys.add('courseCode')
    else:
        res = {}
    for re in RE_LIST:
        m = match(re, title)
        if m is not None:
            res.update(m.groupdict())
            break
    if drop_keys:
        for key in res.keys():
            if key in drop_keys:
                del res[key]
    if not res.get('name'):
        res['name'] = title.strip()

    return res


if __name__ == '__main__':
    # Standard library:
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*- vim: ts=8 sts=4 sw=4 si et tw=79
"""
visaplan.plone.tools.cfg: Zugriff auf "Produkt"konfiguration ...

... in buildout-Skript:

    zope-conf-additional +=
        <product-config name>
            debug_active 1
        </product-config>

(kopiert nach ../../../parts/instance/etc/zope.conf, führender Leerraum wird
entfernt)

ACHTUNG - Der Defaultwert für get_debug_active muß ein String sein; z. B.:

    from Globals import DevelopmentMode
    from visaplan.plone.tools.cfg import get_debug_active
    debug_active = get_debug_active('tree', str(DevelopmentMode))

Siehe auch:

- visaplan.tools.dicts.getOption

sowie die als Factory (zur Konversion der Strings zu den benötigten Datentypen)
vorgesehenen Funktionen

- visaplan.tools.minifuncs.makeBool
- visaplan.tools.minifuncs.NoneOrInt
- ...
- visaplan.tools.lands0.makeListOfStrings
"""

# Python compatibility:
from __future__ import absolute_import, print_function

from six import string_types as six_string_types
from six.moves import map

__all__ = ['get_debug_active',
           'get_raw_config',
           'get_config',
           'split_filename',
           ]

# Standard library:
from collections import defaultdict
from os.path import normpath, sep, splitext
from traceback import extract_stack

# Logging / Debugging:
from logging import getLogger

logger = getLogger('visaplan.plone.tools')

# for convenience/development:
try:
    # Zope:
    from App.config import getConfiguration
except ImportError:
    if __name__ != '__main__':
        raise

# visaplan:
from visaplan.tools.minifuncs import NoneOrString, makeBool


def get_debug_active(product, default=None):
    """
    Lies den Konfigurationswert 'debug_active' für das übergebene Produkt aus.
    Das Ergebnis ist ein Wahrheitswert oder eine Ganzzahl.
    """
    try:
        dic = getConfiguration().product_config.get(product, {})
    except AttributeError as e:
        logger.warn('get_debug_active(%(product)r):'
                    ' Can\'t read configuration!',
                    locals())
        logger.exception(e)
    except KeyError as e:
        logger.warn('get_debug_active(%(product)r):'
                    ' obviously no configuration for this product!',
                    locals())
        logger.exception(e)
    else:
        if 'debug-active' in dic:
            raise ValueError('Fehlerhafte Konfiguration fuer Produkt'
                             ' %(product)r!'
                             % locals())
        val = dic.get('debug_active', '')
        return makeBool(val, default)
    # in case of errors:
    logger.info('get_debug_active(%(product)r):'
                ' using default %(default)r',
                locals())
    return makeBool(default)


def get_raw_config(product=None, defaults={}, fn=None):
    """
    Lies die Konfiguration für das übergebene Produkt aus
    und gib das "rohe" Dictionary zurück (ohne jegliche Wertekonversion).

    Schlüssel, die in der Konfiguration fehlen,
    werden ggf. dem defaults-dict entnommen.

    Siehe auch visaplan.tools.dicts.getOption
    """
    if fn is not None:
        if product is not None:
            raise ValueError('Bitte nur entweder <product> (%(product)r)'
                             ' oder <fn> (%(fn)r) angeben!'
                             % locals())
        product = split_filename(fn)[0]

    conf = getConfiguration()
    if conf is None:
        return defaults
    dic = conf.product_config.get(product, {})
    if defaults:
        res = {}
        res.update(defaults)
        res.update(dic)
        return res
    else:
        return dic


# TODO: optionale Umgebungsvariablen ("veraltend") unterstützen,
#       z. B. W3L_URL (<product>_<key>);
#       siehe pdf/config.py
def get_config(**kwargs):
    """
    Lies die Konfiguration für das übergebene Produkt aus
    und gib ein Dict. mit konvertierten Werten zurück

    Alle Argumente sind semantisch optional und müssen daher sinnvollerweise
    benannt angegeben werden; allerdings wird genau eine Option aus <product>
    und <fn> benötigt (wie für --> get_raw_config), und die Angabe von
    <factories> wird praktisch immer benötigt.

    Optionen:
    fn -- Dateiname, üblicherweise des aufrufenden Moduls (__file__)
    product -- alternativ: Name des konfigurierten Produkts
    defaults -- Vorgabewerte (Dict)
    factories -- Dict. von Funktionen, die aus den Konfigurations- oder, wenn
                 fehlend, Vorgabewerten die zu verwendenen Werte erzeugen

    Alle Schlüssel sollten mindestens in einem der Argumente <factories> oder
    <defaults> explizit sein; im jeweils anderen können sie über den
    defaultdict-Mechanismus kommen.
    """
    if 'fn' in kwargs:
        assert 'product' not in kwargs, (
                'Bitte nur *entweder* fn (%(fn)r)'
                ' *oder* product (%(product)r) angeben!'
                ) % locals()
        fn = kwargs.pop('fn')
        product = split_filename(fn)[0]
    else:
        product = kwargs.pop('product')
    factories = kwargs.pop('factories', None)
    if factories is None:
        factories = defaultdict(NoneOrString)
    dic = kwargs.pop('defaults', {})
    conf = getConfiguration().product_config.get(product, {})
    dic.update(conf)

    fact_keys = set(factories.keys())
    done_keys = set()
    for key, val in dic.items():
        done_keys.add(key)
        if isinstance(val, six_string_types):
            dic[key] = factories[key](val)
    fact_keys.difference_update(done_keys)
    for key in fact_keys:
        # klappt, wenn defaults ein defaultdict ist:
        dic[key] = factories[key](dic[key])
    return dict(dic)


def split_filename(fn, name=None,
                   prefix=None, suffix=None,
                   stack_limit=2):
    """
    Ermittle einen Namen und ein Label aus dem Dateinamen,
    der ggf. aus dem Aufrufstack ermittelt wird.

    Zunächst "richtige" Entwicklungs- oder installierte Quellen:

    >>> split_filename('/opt/zope/instances/unitracc-devel/src/visaplan.plone.tools/src/visaplan/plone/tools/log.py')
    ('visaplan.plone.tools', 'visaplan.plone.tools:log (DEV)')
    >>> split_filename('/opt/zope/common/eggs/visaplan.tools-1.0-py2.7.egg/visaplan/tools/__init__.py')
    ('visaplan.tools', 'visaplan.tools')
    >>> split_filename('/opt/zope/common/eggs/visaplan.tools-1.1-py2.7.egg/src/visaplan/tools/misc.py')
    ('visaplan.tools', 'visaplan.tools:misc')

    Nun die historischen Varianten:

    >>> split_filename('/.../Products/unitracc/browser/tan/browser.py')
    ('tan', 'unitracc@@tan')
    >>> split_filename('/.../Products/zkb/browser/projectreview/browser.py')
    ('projectreview', 'zkb@@projectreview')
    >>> split_filename('/.../Products/unitracc/adapters/sqlwrapper/adapter.py')
    ('sqlwrapper', 'unitracc->sqlwrapper')

    Hilfsmodule:
    >>> split_filename('/.../Products/unitracc/browser/tan/utils.py')
    ('tan', 'unitracc@@tan:utils')

    Sonstiges (Name für Settings noch nicht fix):
    >>> split_filename('/.../Products/unitracc/upgrades/to_20236.py')
    ('to_20236_oder_so', 'unitracc:upgrades:to_20236')
    >>> split_filename('/.../Products/unitracc/tools/misc.py')
    ('misc_oder_so', 'unitracc:tools:misc')

    Die Unterstützung des Aufrufs aus Fremdpaketen ist noch etwas umständlich:
    >>> split_filename('/.../src/visaplan.UnitraccShop/src/visaplan/UnitraccShop/__init__.py',
    ...                name='visaplan.UnitraccShop',
    ...                prefix='')
    ('visaplan.UnitraccShop', 'visaplan.UnitraccShop')

    >>> split_filename('/opt/zope/common/eggs/visaplan.plone.tools-1.5.0-py2.7.egg/visaplan/plone/tools/cfg.py')
    ('visaplan.plone.tools', 'visaplan.plone.tools:cfg')

    Der folgende Aufruf zeitigt noch nicht das gewünschte Ergebnis:
    >>> split_filename('.../src/Products.unitracc/src/Products/unitracc/browser/viewpreview/browser.pyc')
    ... # doctest: +SKIP
    ('Products.unitracc', 'Products.unitracc@@viewpreview (DEV)')
    ... sondern:
    ('Products.unitracc', 'Products.unitracc:browser.viewpreview.browser (DEV)')
    """
    if name:
        if prefix is None:
            prefix = 'unitracc@@'
        if suffix is None:
            suffix = ''
        return (name,
                '%(prefix)s%(name)s%(suffix)s' % locals())

    if fn is None:
        raw_info = extract_stack(limit=stack_limit)
        fn = raw_info[0][0]

    lst = normpath(fn).split(sep)  # alle
    _pkg, _mod, _devel = split_srcfilename(lst)
    if _pkg is not None:
        _label = _pkg
        if _mod is not None:
            _label += ':'+_mod
        if _devel:
            _label += ' (DEV)'
        return (_pkg, _label)
    _prefix = None
    _name = None
    _suffix = None
    _fn = lst.pop()
    lst2 = list(lst)
    _fn_base, _fn_ext = splitext(_fn)
    _dn = lst.pop()
    _pn = lst.pop()  # parent name ('browser')
    _Pn = lst.pop()  # Product name ('unitracc')
    if _fn_base == 'adapter':
        _prefix = _Pn + '->'
        _name = _dn
        _suffix = ''
    elif _fn_base == 'browser':
        _prefix = _Pn + '@@'
        _name = _dn
        _suffix = ''
    elif _pn.startswith('browser'):
        _prefix = _Pn + '@@'
        _name = _dn
        _suffix = ':'+_fn_base
    elif _pn.startswith('adapter'):
        _prefix = _Pn + '->'
        _name = _dn
        _suffix = ':'+_fn_base
    elif 'Products' in lst2[:-1]:
        # "historic" components etc. from our monolith:
        pi = lst2.index('Products')
        return (_fn_base+'_oder_so',
                ':'.join(lst2[pi+1:] + [_fn_base]),
                )
    else:
        ohwell = '.'.join(lst2) or 'oh well'
        return (ohwell, ohwell)
        raw = None
        pp(_fn_base=_fn_base, _dn=_dn, lst2=lst2, raw=raw)
        set_trace()
        res = None
        return res

        try:
            pi = lst2.index('Products')
            return (_fn_base+'_oder_so',
                    ':'.join(lst2[pi+1:] + [_fn_base]),
                    )
        except ValueError as e:
            print(str(e))
            return '?'.join(lst2)
            raise ValueError('Unsupported filename: %(fn)r' % locals())
    if prefix is None:
        prefix = _prefix
    if name is None:
        name = _name
    if suffix is None:
        suffix = _suffix
    return (name,
            '%(prefix)s%(name)s%(suffix)s' % locals())


def parse_egg_dirname(dn):
    """
    Extrahiere die folgenden Informationen aus dem Namen:

    - den Paketnamen
    - die Versionsangabe
    - das Python-Versionstupel

    >>> parse_egg_dirname('visaplan.tools-1.0-py2.7.egg')
    ('visaplan.tools', '1.0', (2, 7))

    Sonstige Namen werden nicht erkannt:
    >>> parse_egg_dirname('visaplan.tools')
    (None, None, None)
    """
    if not dn.endswith('.egg'):
        return (None, None, None)
    lapypo = dn.rfind('-py')
    if lapypo == -1:
        return (None, None, None)
    endofpa = dn[:lapypo].rfind('-')
    if endofpa == -1:
        return (None, None, None)
    pa = dn[:endofpa]
    pv = dn[endofpa+1:lapypo]
    pyraw = dn[lapypo+3:-4]
    try:
        pytup = tuple(map(int, pyraw.split('.')))
    except ValueError:
        return (None, None, None)
    else:
        return pa, pv, pytup


def _pkg_and_submodule(pkg, tail):
    """
    >>> _pkg_and_submodule('visaplan.tools', ['visaplan', 'tools', 'coding.py'])
    ('visaplan.tools', 'coding')
    >>> _pkg_and_submodule(['visaplan', 'tools'], ['visaplan', 'tools', '__init__.py'])
    ('visaplan.tools', None)
    """
    if isinstance(pkg, six_string_types):
        pkg_as_list = pkg.split('.')
    else:
        pkg_as_list, pkg = pkg, '.'.join(pkg)
    pkg_len = len(pkg_as_list)
    if tail[:pkg_len] != pkg_as_list:
        raise ValueError('tail %(tail)r mismatches package %(pkg)r!'
                         % locals())
    if not tail[pkg_len:]:
        return pkg, None
    modfile = tail.pop()
    module, ext = splitext(modfile)
    if module == '__init__':
        return pkg, '.'.join(tail[pkg_len:]) or None
    return pkg, '.'.join(tail[pkg_len:]+[module])


def split_srcfilename(fn):
    """
    Nimm den Namen einer Python-Datei entgegen und extrahiere hieraus drei
    Informationen:
    - das Paket
    - das Modul
    - handelt es sich um eine Entwicklungsressource?

    >>> split_srcfilename('/opt/zope/common/eggs/visaplan.tools-1.0-py2.7.egg/visaplan/tools/coding.py')
    ('visaplan.tools', 'coding', False)
    >>> split_srcfilename('/opt/zope/common/eggs/visaplan.tools-1.1-py2.7.egg/src/visaplan/tools/misc.py')
    ('visaplan.tools', 'misc', False)
    >>> split_srcfilename('/opt/zope/common/eggs/visaplan.tools-1.0-py2.7.egg/visaplan/tools/__init__.py')
    ('visaplan.tools', None, False)
    >>> split_srcfilename('/opt/zope/instances/unitracc-devel/src/visaplan.plone.tools/src/visaplan/plone/tools/log.py')
    ('visaplan.plone.tools', 'log', True)
    >>> split_srcfilename('.../src/visaplan.tools/src/visaplan/tools/debug.py')
    ('visaplan.tools', 'debug', True)
    >>> split_srcfilename('.../src/visaplan.tools/visaplan/tools/coding.py')
    ('visaplan.tools', 'coding', True)

    Fallback-Wert für Pfade ohne src- eder egg-Komponente (kein Paket erkannt):

    >>> split_srcfilename('/some/strange/unexpected/path.py')
    (None, '.../strange/unexpected/path.py', True)

    Der folgende Aufruf zeitigt noch nicht das gewünschte Ergebnis:
    >>> split_srcfilename('.../src/Products.unitracc/src/Products/unitracc/browser/viewpreview/utils.pyc')
    ...   # doctest: +SKIP
    """
    if isinstance(fn, six_string_types):
        fn_list = fn.split(sep)
    else:
        fn_list = list(fn)
    i = 0
    src_positions = []
    package_name = None
    for chunk in fn_list:
        if chunk == 'src':
            src_positions.append(i)
        else:
            _pn = parse_egg_dirname(chunk)[0]
            if _pn is not None:
                package_name = _pn
                i += 1
                break
        i += 1

    if package_name is not None:
        if fn_list[i] == 'src':
            i += 1
        return _pkg_and_submodule(package_name, fn_list[i:]) + (False,)

    # kein Egg;
    src_bases = src_positions[-2:]
    if src_bases[1:]:  # ja, es gibt (mindestens) zwei:
        sb1, sb2 = src_bases
        if sb2 <= sb1 + 1:
            raise ValueError('.../src/src/... ?! (%(sb1)d, %(sb2)d)'
                             % locals())
        return _pkg_and_submodule(fn_list[sb1+1], fn_list[sb2+1:]) + (True,)
    elif src_bases:
        sb1 = src_bases[0]
        return _pkg_and_submodule(fn_list[sb1+1], fn_list[sb1+2:]) + (True,)
    return (None,
            sep.join(['...']
                     +fn_list[-3:]
                     ),
            True)


if __name__ == '__main__':
  if 0:
      s = '/opt/zope/common/eggs/visaplan.plone.tools-1.5.0-py2.7.egg/visaplan/plone/tools/cfg.py'
      s = '.../src/Products.unitracc/src/Products/unitracc/browser/viewpreview/browser.pyc'
      # Logging / Debugging:
      from pdb import set_trace
      from visaplan.tools.debug import pp
      pp(s)
      set_trace()
      res = split_srcfilename(s)
      print(res)
  else:
    # Standard library:
    import doctest
    doctest.testmod()

# Logging / Debugging:
from pdb import set_trace
from visaplan.tools.debug import pp



# -*- coding: utf-8 -*- äöü vim: ts=8 sts=4 sw=4 si et tw=79
"""\
Tools for "brains"
"""

# Python compatibility:
from __future__ import absolute_import

from six import string_types as six_string_types
from six.moves import map

__author__ = "Tobias Herp <tobias.herp@visaplan.com>"
VERSION = (0,
           6,  # aufgeräumt
           )
__version__ = '.'.join(map(str, VERSION))


__all__ = [
           'make_collector',
           ]


def make_collector(use, all=None, any=None, separator=" ", empty=None):
    """
    Gib eine Funktion zurück, die aus einem Brain-Objekt die angegebenen
    Attribute extrahiert und als String (Normalfall)
    oder als Liste (separator == None) zurückgibt.
    Argumente:

    use -- geordnete Sequenz der Attributnamen.
           Jedes Element der Sequenz darf ein 2-Tupel sein;
           in diesem Fall ist der erste Teil der Attributname, der zweite
           eine Transformationsfunktion, der der Wert übergeben wird.
    all -- wenn übergeben, müssen alle hier angegebenen Attribute einen
           nicht-leeren Wert haben; andernfalls wird <empty> zurückgegeben
    any -- komplementär zu <all>: wenn ein beliebiges der hier angegebenen
           Attribute einen nicht-leeren Wert hat ...
    separator -- Wenn None, wird ggf. eine Liste zurückgegeben; wenn das all-
                 oder any-Kriterium nicht erfüllt ist, eine leere Liste
    empty -- der Wert, wenn das all- oder any-Kriterium nicht erfüllt ist.
             Wird ignoriert, wenn <separator> None ist (was bedeutet, daß eine
             Liste zurückgegeben wird; in diesem Fall kommt eine leere Liste
             zurück).

    Die Angaben für all bzw. any sollten Untermengen von use sein - ansonsten
    wird hier nicht versprochen, daß sie korrekt verarbeitet werden!

    Derzeit kann nur *entweder* all *oder* any angegeben werden;
    für genau ein enthaltenes Element ist die Bedeutung allerdings identisch.

    >>> from sys import path
    >>> from os.path import dirname
    >>> path.insert(0, dirname(__file__))
    >>> from mock import MockBrain
    >>> caesar=MockBrain(getAcademicTitle='Dr.', getFirstname='Hase',
    ...                  getLastname='Caesar')
    >>> use=sorted(caesar.keys())
    >>> f_any=make_collector(use=use)
    >>> otto=MockBrain(getFirstname='Otto')

    Die Funktion f_any gibt nun immer einen String zurück, wenn das übergebene
    dict für mindestens eines dieser Attribute einen Wert hat.

    >>> f_any(caesar)
    'Dr. Hase Caesar'
    >>> f_any(otto)
    'Otto'

    Um nur dann etwas auszugeben, wenn ein Nachname vorhanden ist
    (ein einzelner Wert darf als String übergeben werden):

    >>> f_all=make_collector(use=use, all='getLastname')
    >>> f_all(caesar)
    'Dr. Hase Caesar'
    >>> f_all(otto)

    Achtung:
    Die Collectorfunktionen können (und sollten möglichst) schon beim Laden des
    Moduls erzeugt werden; es sind nämlich noch nicht alle möglichen Varianten
    implementiert ...
    """

    # ----------------------- [ make_collector: Argumente prüfen ... [
    def make_set(sos):
        """
        sos -- string or sequence

        Gib ein Set zurück.
        Wenn ein String übergeben wurde, füge ihn als Ganzes hinzu:

        >>> make_set('getFirstname')
        set(['getFirstname'])

        Sonstige Sequenzen und Sets werden normal verarbeitet:

        >>> make_set(['getFirstname', 'getLastname'])
        set(['getFirstname', 'getLastname'])
        """
        if isinstance(sos, six_string_types):
            return set([sos])
        else:
            return set(sos)

    assert use
    if all is None:
        if any is None:
            any = make_set(use)
        else:
            any = make_set(any)
    else:
        assert any is None, ('Es kann nur *entweder* all (%(all)r) *oder*'
                             ' any (%(any)r) angegeben werden!'
                             % locals())
        all = make_set(all)
    if separator is None:
        assert empty is None, 'empty wird nicht verwendet!'
    has_tuples = 0
    for key in use:
        if isinstance(key, tuple):
            has_tuples = 1
            break
    if has_tuples:
        use2 = []
        for key in use:
            if isinstance(key, tuple):
                use2.append(key)
            else:
                use2.append((key, None))
        use = use2
    # ----------------------- ] ... make_collector: Argumente prüfen ]

    # --------------------------------- [ einfache Werte, String ... [
    def collect_any(o):
        """
        any angegeben, ggf. empty
        """
        found = set()
        res = []
        allnames = dir(o)
        for key in use:
            if key in o:
                val = getattr(o, key, None)
                if val:
                    res.append(val)
                    found.add(key)
        if res and found.intersection(any):
            return separator.join(res)
        else:
            return empty

    def collect_all(o):
        """
        all angegeben, ggf. empty
        """
        res = []
        allnames = dir(o)
        for key in use:
            if key in o:
                val = getattr(o, key, None)
                if val:
                    res.append(val)
                    continue
            if key in all:
                return empty
        if res:
            return separator.join(res)
        else:
            return empty
    # --------------------------------- ] ... einfache Werte, String ]

    # ---------------------------- [ einzelne Funktionen, String ... [
    def collect_any_funcs(o):
        """
        any angegeben, ggf. empty
        """
        found = set()
        res = []
        allnames = dir(o)
        for (key, func) in use:
            if key in o:
                val = getattr(o, key, None)
                if func is not None:
                    val = func(val)
                if val:
                    res.append(val)
                    found.add(key)
        if res and found.intersection(any):
            return separator.join(res)
        else:
            return empty

    def collect_all_funcs(o):
        """
        all angegeben, ggf. empty
        """
        res = []
        allnames = dir(o)
        for (key, func) in use:
            if key in o:
                val = getattr(o, key, None)
                if func is not None:
                    val = func(val)
                if val:
                    res.append(val)
                    continue
            if key in all:
                return empty
        if res:
            return separator.join(res)
        else:
            return empty
    # ---------------------------- ] ... einzelne Funktionen, String ]

    # ---------------------------------- [ einfache Werte, Liste ... [
    def collect_any_list(o):
        """
        any angegeben, ggf. []
        """
        found = set()
        res = []
        allnames = dir(o)
        for key in use:
            if key in o:
                val = getattr(o, key, None)
                if val:
                    res.append(val)
                    found.add(key)
        if res and found.intersection(any):
            return separator.join(res)
        else:
            return []

    def collect_all_list(o):
        """
        all angegeben, ggf. []
        """
        res = []
        allnames = dir(o)
        for key in use:
            if key in o:
                val = getattr(o, key, None)
                if val:
                    res.append(val)
                    continue
            if key in all:
                return []
        if res:
            return separator.join(res)
        else:
            return []
    # ---------------------------------- ] ... einfache Werte, Liste ]

    if separator is not None:
        if any:
            if has_tuples:
                return collect_any_funcs
            else:
                return collect_any
        else:
            assert all
            if has_tuples:
                return collect_all_funcs
            else:
                return collect_all
    ## -------------------------------------------- ... make_collector


if __name__ == '__main__':
    # Standard library:
    import doctest
    doctest.testmod()



# -*- coding: utf-8 -*- äöü vim: ts=8 sts=4 sw=4 si et hls tw=79
# Python compatibility:
from __future__ import absolute_import, print_function

from six import text_type as six_text_type

# Standard library:
from collections import defaultdict

# Zope:
from Products.CMFCore.utils import getToolByName

# visaplan:
from visaplan.tools.minifuncs import gimme_None

# Local imports:
from ._data import PRETTY_MASK
from ._group_pio import _parse_init_options
# since this function uses some quite application specific data,
# it is a hot candidate for an initialiation option:
from ._helpers import split_group_id
from visaplan.plone.tools.context import getbrain, make_translator

try:
    # visaplan:
    from visaplan.tools.coding import safe_decode
except ImportError:
    if __name__ != '__main__':
        raise
    def safe_decode(s):
        if isinstance(s, six_text_type):
            return s
        return s.decode('utf-8')

try:
    # visaplan:
    from visaplan.plone.groups.unitraccgroups.utils import pretty_group_title
except ImportError:
    if __name__ != '__main__':
        raise
    print("E: Couldn't import pretty_group_title!")
    pretty_group_title = None


__all__ = [
    'groupinfo_factory',
    ]


# -------------------------------- [ groupinfo_factory ... [
def groupinfo_factory(context, *args, **kwargs):
    """
    Factory-Funktion; als Ersatz für get_group_info_by_id, insbesondere für
    wiederholte Aufrufe.  Indirektionen etc. werden einmalig beim Erstellen der
    Funktion aufgelöst und anschließend in einer Closure vorgehalten.

    Historisch wurden die Optionen pretty, forlist und searchtext auch
    (in dieser Reihenfolge) positional akzeptiert; dies wird nicht mehr empfohlen
    (wenn auch durch _parse_init_options derzeit noch unterstützt):

    pretty -- Gruppennamen auflösen, wenn Suffix an ID und Titel
    forlist -- Minimale Rückgabe (nur ID und Titel),
               aber mit pretty kombinierbar
    missing -- return a dict for missing groups as well,
               and add an `exists` key

    searchtext -- Erzeuge eine Funktion, um einen Suchtext zu erzeugen;
                  diese nimmt keinee Gruppen-ID entgegen (wie die andernfalls
                  erzeugten Funktionen), sondern ein von diesen erzeugtes
                  Info-Dict.
    """
    _parse_init_options(kwargs, args)
    pretty = kwargs['pretty']
    if pretty_group_title is None and pretty:
        raise ValueError('pretty=%(pretty)r not supported; '
                "Couldn't import from visaplan.plone.groups!"
                % kwargs)
    forlist = kwargs['forlist']
    missing = kwargs['missing']
    missing_mask = kwargs.get('missing_group_mask') or None
    searchtext = kwargs['searchtext']

    acl = getToolByName(context, 'acl_users')
    pg = getToolByName(context, 'portal_groups')
    get_group = pg.getGroupById
    translate = make_translator(context)
    GROUPS = acl.source_groups._groups
    if missing:
        missing_mask = translate(missing_mask)

    def minimal_group_info(group_id):
        """
        Return a dict with keys 'id', 'group_title', or empty.
        """
        group = get_group(group_id)

        try:
            thegroup = GROUPS[group_id]
        except KeyError:
            if missing:
                return defaultdict(gimme_None,
                                   id=group_id,
                                   exists=0)
            return {}
        else:
            dict_ = {
                'id': group_id,
                'group_title': safe_decode(thegroup['title']),
                }
            if missing:
                dict_['exists'] = 1
            return dict_

    def basic_group_info(group_id):
        """
        Gib ein Dict. zurück;
        - immer vorhandene Schlüssel:
          id, group_title, group_description, group_manager, group_desktop
        - nur bei automatisch erzeugten Gruppen: role, role_translation, brain

        Argumente:
        group_id -- ein String, normalerweise mit 'group_' beginnend
        """

        group = get_group(group_id)

        try:
            thegroup = GROUPS[group_id]
        except KeyError:
            if missing:
                return defaultdict(gimme_None,
                                   id=group_id,
                                   exists=0)
            return {}
        else:
            dict_ = {
                'id': group_id,
                'group_title': safe_decode(thegroup['title']),
                }
            if missing:
                dict_['exists'] = 1

        # propietary keys:
        dict_['group_description'] = safe_decode(thegroup['description'])
        dict_['group_manager'] = thegroup.get('group_manager')
        dict_['group_desktop'] = thegroup.get('group_desktop')

        dic = split_group_id(group_id)
        if dic['role'] is not None:
            dict_.update(dic)

        # refered object (for <group_><uid>[_role]:
        dict_['role_translation'] = translate(dic['role'])  # local role-to-be-mapped
        dict_['brain'] = getbrain(context, dic['uid'])  # refered object
        return dict_

    def pretty_group_info(group_id):
        """
        Ruft basic_group_info auf und fügt einen Schlüssel 'pretty_title'
        hinzu, der den Gruppentitel ohne das Rollensuffix enthält.
        """
        dic = basic_group_info(group_id)
        if not dic:
            assert not missing
            return dic
        elif not dic.get('exists', 1):
            assert missing
            if 'group_id' in missing_mask:
                dic['group_id'] = group_id
            dic['pretty_title'] = missing_mask.format(**dic)
            return dic

        if 'role' not in dic:
            dic['pretty_title'] = translate(dic['group_title'])
            return dic
        liz = dic['group_title'].split()
        if liz and liz[-1] == dic['role']:
            stem = u' '.join(liz[:-1])
            mask = PRETTY_MASK[dic['role']]
            dic['pretty_title'] = translate(mask).format(group=stem)
        else:
            dic['pretty_title'] = translate(dic['group_title'])
        return dic

    def minimal2_group_info(group_id):
        """
        Ruft minimal_group_info auf und modifiziert ggf. den Schlüssel
        'group_title' (entsprechend dem von pretty_group_info zurückgegebenen
        Schlüssel 'pretty_title')
        """
        dic = minimal_group_info(group_id)
        if not dic:
            assert not missing
            return dic
        elif not dic.get('exists', 1):
            assert missing
            return dic

        dic2 = split_group_id(group_id)
        if dic2['role'] is not None:
            dic.update(dic2)
        if 'role' not in dic:
            dic['group_title'] = translate(dic['group_title'])
            return dic
        liz = dic['group_title'].split()
        if liz and liz[-1] == six_text_type(dic['role']):
            stem = u' '.join(liz[:-1])
            mask = PRETTY_MASK[dic['role']]
            dic['group_title'] = translate(mask).format(group=stem)
        else:
            dic['group_title'] = translate(dic['group_title'])
        return dic

    def make_searchstring(group_info):
        """
        Arbeite direkt auf einem group_info-Dict;
        Gib kein Dict zurück, sondern einen String für Suchzwecke
        """
        try:
            group_id = group_info['id']
        except KeyError:
            raise
        try:
            res = [safe_decode(group_info['title']),
                   safe_decode(group_id)]
        except KeyError:
            print(list(group_info.items()))
            raise
            return u''
        dic2 = split_group_id(group_id)
        prettify = True
        for val in dic2.values():
            if val is None:
                prettify = False
                break  # es sind immer alle None, oder alle nicht None
            else:
                res.append(safe_decode(val))
        if prettify:
            pretty = pretty_group_title(group_id, res[0], translate)
            if pretty is not None:
                res.append(safe_decode(pretty))
        descr = group_info['description']
        if descr:
            res.append(safe_decode(descr))
        return u' '.join(res)

    if searchtext:
        return make_searchstring
    if forlist:
        if pretty:
            return minimal2_group_info
        else:
            return minimal_group_info
    if pretty:
        return pretty_group_info
    else:
        return basic_group_info
# -------------------------------- ] ... groupinfo_factory ]



# -*- coding: utf-8 -*- äöü vim: sw=4 sts=4 et tw=79
"""
Tools für Produkt-Setup (Migrationsschritte, "upgrade steps"): _tree
"""

# Python compatibility:
from __future__ import absolute_import

# Standard library:
from posixpath import normpath
from string import capitalize

# Zope:
from Products.CMFCore.utils import getToolByName

# Local imports:
from visaplan.plone.tools.setup._args import (
    _extract_move_args,
    extract_layout_switch,
    extract_menu_switch,
    )
from visaplan.plone.tools.setup._misc import _traversable_path, make_title
from visaplan.plone.tools.setup._o_tools import (
    handle_language,
    handle_layout,
    handle_menu,
    handle_title,
    make_notes_logger,
    )
from visaplan.plone.tools.setup._reindex import make_reindexer

# Logging / Debugging:
import logging

__all__ = [
        'make_subfolder_creator',
        ]


# see also _get_object.py: make_object_getter
def make_subfolder_creator(**kwargs):
    """
    Erzeuge eine Funktion, die einen Ordner erzeugt und zurückgibt.
    Optionen, alle benannt zu übergeben:

    logger - der zu verwendende Logger
    title_factory - Funktion, um einen ggf. fehlenden Title zu erzeugen
    reindexer - wie von --> make_reindexer erzeugt

    Vorgabewerte für die zu erzeugende Funktion:

    set_menu - Soll die Aktivierung von Menüeinträgen umgeschaltet werden?
               Siehe ._args.extract_menu_switch.
    parent - das schon existierende ordnerartige Elternobjekt
    view_id - die ID der aufzuschaltenden Ansicht

    Argumene für einen zu erzeugenden Reindexer:

    idxs - eine Liste der zu aktualisierenden Indexe;
           make_reindexer wird per Vorgabe ['getId'] verwenden

    """
    parent = kwargs.pop('parent', None)
    logger = kwargs.pop('logger', None)
    if logger is None:
        logger = logging.getLogger('new folder')
    title_factory = kwargs.pop('title_factory', None)
    if title_factory is None:
        title_factory = make_title
    view_id, set_layout, get_layout = extract_layout_switch(kwargs)
    set_menu = kwargs.pop('set_menu', None)

    reindexer = kwargs.pop('reindexer', None)
    idxs = kwargs.pop('idxs', None)
    if reindexer is None and (set_menu is None or set_menu):
        reindexer = make_reindexer(logger=logger,
                                   context=parent,
                                   idxs=idxs)
    elif reindexer is not None and idxs is not None:
        logger.warn('Ignoring idxs value %(idxs)r', locals())
    reindex = kwargs.pop('reindex', reindexer is not None
                                    or None)
    if reindex and reindexer is None:
        reindexer = make_reindexer(logger=logger,
                                   context=parent)

    def new_folder(id=None, title=None, parent=parent,
                   view_id=view_id,
                   set_layout=set_layout,
                   get_layout=get_layout,
                   reindex=reindex,
                   reindexer=reindexer,
                   set_menu=set_menu,
                   **kwargs):
        created = False
        changed = False
        create = kwargs.pop('create', None)
        path = kwargs.pop('path', None)
        if create is None:
            create = bool(id or path)
        elif create:
            if not (id or path):
                raise TypeError('At least "id" or "path" is needed!')

        if id is None:
            if path is not None:
                logger.warn('Create or change %(path)r: '
                            'ignoring parent %(parent)r!',
                            locals())
                portal = getToolByName(parent, 'portal_url').getPortalObject()
                try:
                    new_child = portal.restrictedTraverse(_traversable_path(path))
                except KeyError:
                    logger.info('Object %(portal)r[%(path)r] not (yet) found!', locals())
                    new_child = None
        else:
            if path is not None:
                logger.warn('Looking for %(id)r in %(parent)r; '
                            'ignoring path %(path)r!',
                            locals())
            new_child = getattr(parent, id, None)

        if new_child is None and not create:
            logger.info('object not found (id=%(id)r, path=%(path)r, parent=%(parent)r', locals())
            return None

        if new_child is None:
            if id is None:  # in this case we have a path!
                chunks = normpath(path).split('/')
                id = chunks[-1]
                if not id:
                    raise ValueError('Empty id, and not really helpful path'
                            ' value! (%(path)r)'
                            % locals())
                parent = '/'.join(chunks[:-1])
                if parent in ('', '/'):
                    parent = portal
                else:
                    parent = portal.restrictedTraverse(_traversable_path(parent))

            if title is None:
                title = title_factory(id)
                logger.info('Default title for %(id)s is %(title)r', locals())
            logger.info('Creating folder %(parent)r/%(id)s (%(title)s)...',
                        locals())
            parent.manage_addFolder(id=id, title=title)
            new_child = getattr(parent, id)
            created = True
        elif title is not None:
            logger.info('Ignoring new title %(title)r for existing %(new_child)r',
                        locals())

        lognotes = make_notes_logger(logger)

        # ---------- [set_]language, [set_]canonical:
        ch, notes = handle_language(new_child, kwargs, created)
        # we don't have a changes counter here yet
        for tup in notes:
            lognotes(tup)
        if ch:
            changed = True

        # ---------- [set_]layout, view_id:
        if view_id is not None:
            kwargs['view_id'] = view_id
        kwargs.update(set_layout=set_layout, get_layout=get_layout)
        ch, notes, upd = handle_layout(new_child, kwargs, created)
        for tup in notes:
            lognotes(tup)
        if ch:
            changed = True

        # ---------- [set_]title:
        if not created:  # in this case we *have* set a title already
            if title is None:
                kwargs.update(default_title=title_factory(id))
            else:
                kwargs.update(title=title)
            ch, notes = handle_title(new_child, kwargs, created)
            for tup in notes:
                lognotes(tup)
            if ch:
                changed = True

        switch_menu = extract_menu_switch(kwargs, created)
        if switch_menu is not None:
            # see new function ._o_tools.handle_menu
            val = not switch_menu
            new_child.setExcludeFromNav(val)
            logger.info('%(new_child)r: setExcludeFromNav(%(val)r)', locals())
            changed = True  # TODO: use handle_menu

        uid = kwargs.pop('uid', None)
        if uid:
            logger.info('%(new_child)r: setting UID to %(uid)r', locals())
            new_child._setUID(uid)

        if reindex is None:
            reindex = (created
                       or changed
                       or reindexer is not None)
        if reindex and reindexer is None:
            reindexer = make_reindexer(logger=logger,
                                       context=parent)
        if reindex:
            res = reindexer(o=new_child)
            logger.info('%(res)r <-- reindexed %(new_child)r', locals())
        else:
            logger.info('NOT reindexed: %(new_child)r', locals())

        if kwargs:
            logger.warn('unused Arguments: %(kwargs)s', locals())
        return new_child

    return new_folder



# -*- coding: utf-8 -*- äöü vim: sw=4 sts=4 et tw=79
"""
Tools für Produkt-Setup (Migrationsschritte, "upgrade steps"): _watch
"""

# Python compatibility:
from __future__ import absolute_import

from six import string_types as six_string_types

# Standard library:
from collections import defaultdict

__all__ = [
        'make_watcher_function',
        ]


def make_watcher_function(val, logger, **kwargs):
    """
    Erzeuge eine Funktion (für Debugging-Zwecke), die zwei Argumente
    entgegennimmt - einen Schlüssel (etwa eine UID) und einen String, z. B.
    einen Workflow-Zielstatus - und einen Wahrheitswert zurückgibt.
    Der aufrufende Code kann dann den Debugger aufrufen.

    Für Abkürzung der Doctests:
    >>> def mwf(val): return make_watcher_function(val, logger=None)

    Mögliche Werte:
    Ein 2-Tupel (key, val):
    >>> f = mwf(('abc123', 'restricted'))

    Eine Liste von 2-Tupeln:
    >>> f = mwf([('abc123', 'restricted'), ('cde465', 'restricted'))

    Dies kann auch abgekürzt werden (gleichwertige Darstellungen):
    >>> f = mwf([(('abc123', 'cde465a'),  'restricted')])
    >>> f = mwf( (('abc123', 'cde465a'),  'restricted') )
    >>> f = mwf([(('abc123', 'cde465a'), ['restricted'])])

    Als Dict (nah an der internen Darstellung; wiederum gleichwertig):
    >>> f = mwf({'abc123': 'restricted', 'cde465': 'restricted'})
    >>> f = mwf({('abc123', 'cde465'): ('restricted',)})

    Die zurückgegebene Funktion verwendet die Argumentnamen 'key' und 'val',
    die in der <msg> verwendet werden können; wenn ein Logger != None übergeben
    wurde, wird eine entsprechende Meldung protokolliert.
    """
    lists_for_uid = defaultdict(list)
    msg = kwargs.pop('msg', 'WATCHED CASE: %(key)r --> %(val)r')
    all_value = kwargs.pop('all_value', 'ALL')
    def checked_value(thelist):
        if (thelist != all_value and
            isinstance(thelist, six_string_types)
            ):
            return [thelist]
        if isinstance(thelist, tuple):
            return list(thelist)
        return thelist

    def nonstring_sequence(val):
        if isinstance(val, (list, tuple)):
            return val
        return [val]

    if isinstance(val, tuple):
        keys, thelist = tup
        for key in nonstring_sequence(keys):
            lists_for_uid[key] = checked_value(thelist)
    elif isinstance(val, list):
        for tup in val:
            keys, thelist = tup
            for key in nonstring_sequence(keys):
                lists_for_uid[key].extend(checked_value(thelist))
    elif isinstance(val, dict):
        for keys, thelist in val.items():
            for key in nonstring_sequence(keys):
                lists_for_uid[key] = checked_value(thelist)

    def watch(key, val):
        tmp = lists_for_uid[key]
        if tmp == all_value:
            if logger is not None:
                logger.info(msg, locals())
            return True
        elif isinstance(tmp, list):
            res = val in tmp
            if res and logger is not None:
                logger.info(msg, locals())
            return res
        else:
            return False

    return watch



# -*- coding: utf-8 -*- vim: ts=8 sts=4 sw=4 si et tw=79
"""
visaplan.plone.tools._at2 -- imported via .attools
"""

# Python compatibility:
from __future__ import absolute_import, print_function

from six import string_types as six_string_types

# Standard library:
from collections import defaultdict

# visaplan:
from visaplan.tools.lands0 import makeSet

# "Log-Level" für generierte Funktion:
BASE, CHANGE, NOP, SKIP, DEBUG = (1, 2, 3, 4, 99)

CORE_FIELDS = set([
        'id', 'title',
        ])
BOOLEAN_FIELDS = set([
        'excludeFromNav',
        'allowDiscussion',
        'excludeFromSearch',
        ])
DATE_FIELDS = set([
        'creation_date', 'modification_date',
        'effectiveDate', 'expirationDate',
        ])
UNITRACC_NUMERIC_FIELDS = set([
        'price_shop',
        'shopduration',
        ])
UNITRACC_CHOICES_FIELDS = set([
        'code',
        'unitraccGroups',
        'subPortals',
        ])
PLONE_OTHER_FIELDS = set([  # sonstige (Plone):
        'creators',
        'language',
        ])
UNITRACC_OTHER_FIELDS = set([
        # XML:
        'xml',
        # evtl. später doch berücksichtigen :
        'certificate',
        'contributors',
        'difficulty',
        'duration',
        'html_left',
        'html_right',
        'location',
        'price',
        'relatedItems',
        'subject',
        'target_group',
        ])
# Nicht nach HTML konvertieren:
DEFAULT_BLACKLIST = set()
DEFAULT_BLACKLIST.update(CORE_FIELDS)
DEFAULT_BLACKLIST.update(BOOLEAN_FIELDS)
DEFAULT_BLACKLIST.update(DATE_FIELDS)
DEFAULT_BLACKLIST.update(UNITRACC_NUMERIC_FIELDS)
DEFAULT_BLACKLIST.update(UNITRACC_CHOICES_FIELDS)
DEFAULT_BLACKLIST.update(PLONE_OTHER_FIELDS)
DEFAULT_BLACKLIST.update(UNITRACC_OTHER_FIELDS)


def make_minilogger(verbose, logger, prefix=None):
    """
    Erzeuge den Minilogger, üblicherweise tell(LEVEL, text[, mapping ...])
    """
    def make_mapping(*args):
        if not args:
            return {}
        elif (not args[1:]) and isinstance(args[0], dict):
            return args[0]
        else:
            return args

    if verbose:
        if logger is None:
            if prefix:
                MASK = prefix.replace('%', '%%') + ': %s'
                def _tell(s, *args):
                    print(MASK % (s % make_mapping(*args)))
            else:
                def _tell(s, *args):
                    print(s % make_mapping(*args))
        else:
            _tell = logger.info
        def tell(level, s, *args):
            if verbose >= level:
                _tell(s, *args)
                return True
    else:
        def tell(level, s, *args):
            pass
    return tell


def make_skip_function(tell, **kwargs):
    """
    Gib eine skip-Funktion zurück, die Attributnamen mit der black- oder
    whitelist vergleicht
    """
    if ('sets' in kwargs
        and 'whitelist' not in kwargs
        and 'blacklist' not in kwargs
        ):
        sets = kwargs.pop('sets')
        if sets:
            def skip(fieldname):
                return fieldname not in sets
            return skip

    whitelist = kwargs.pop('whitelist', None)
    blacklist = kwargs.pop('blacklist', None)
    if whitelist is not None:
        assert blacklist is None
        if isinstance(whitelist, six_string_types):
            whitelist = [whitelist]
        whitelist = set(whitelist)
        def skip(fieldname):
            if fieldname not in whitelist:
                tell(SKIP, '%(fieldname)r not in whitelist --> SKIP', locals())
                return True
            return False
    else:
        if blacklist is None:
            blacklist = []
        elif isinstance(blacklist, six_string_types):
            blacklist = [blacklist]
        blacklist = set(blacklist)
        blacklist.update(DEFAULT_BLACKLIST)
        def skip(fieldname):
            if fieldname in blacklist:
                tell(SKIP, '%(fieldname)r in blacklist --> SKIP', locals())
                return True
            return False
    return skip


def make_mimetype_fixer(**kwargs):
    """
    Erzeuge eine Funktion, die das übergebene Archetypes-Objekt korrigiert.

    Alle Optionen müssen als Schlüsselwortargument übergeben werden:

    from_mimetype  -- Vorgabe: keine Filterung
    to_mimetype -- Vorgabe: 'text/html'
    default_output_type -- Vorgabe: 'text/x-html-safe'

    counter -- üblicherweise ein defaultdict(int).  Wenn nicht übergeben, wird
               eines erzeugt, dessen Werte dann nicht zugänglich sind.
    verbose -- ausführliche Ausgaben/Logging?
    logger -- der zu verwendende Logger
    """
    # --------------------------- [ Argumente für die Ausführung ... [
    verbose = kwargs.pop('verbose', None)
    logger = kwargs.pop('logger', None)
    if verbose is None:
        verbose = logger is not None
    tell = make_minilogger(verbose, logger, 'fix_types')

    if 'counter' in kwargs:
        counter = kwargs.pop('counter')
    else:
        counter = defaultdict(int)

    # zu bearbeitende Attribute:
    skip = make_skip_function(tell, **kwargs)
    # --------------------------- ] ... Argumente für die Ausführung ]

    # --------------------------- [ Argumente für die Konversion ... [
    from_mimetypes = []
    from_mimetype = kwargs.pop('from_mimetype', None)
    if from_mimetype is None:
        from_mimetypes = None
    elif isinstance(from_mimetype, six_string_types):
        from_mimetypes.append(from_mimetype)
    else:
        from_mimetypes.extend(list(from_mimetype))

    to_mimetype = kwargs.pop('to_mimetype', 'text/html')
    assert to_mimetype, ('to_mimetype wird benoetigt! (%(to_mimetype)r)'
                         ) % locals()

    if 'default_output_type' in kwargs:
        default_output_type = kwargs.pop('default_output_type')
    elif to_mimetype == 'text/html':
        default_output_type = 'text/x-html-safe'
    else:
        default_output_type = None
    # --------------------------- ] ... Argumente für die Konversion ]

    assert not kwargs, 'Nicht erkannte Argumente: %s' % (kwargs,)

    def fix_types(instance=None, brain=None):
        # Docstring-Generierung: siehe unten
        instance, iinfo = instance_and_label(instance, brain, verbose)
        schema = instance.Schema()
        changed = False
        if verbose:
            tell(BASE, 'fixing types for %(iinfo)s ...', locals())
        for field in schema.values():
            name = field.__name__
            if skip(name):
                continue
            tell(DEBUG, 'field: name=%(name)r,'
                        ' field=%(field)r', locals())
            # nur schreibbare Felder:
            try:
                mutator = field.getMutator(instance)
                if mutator is None:
                    counter[('skipped', name, 'readonly')] += 1
                    tell(SKIP, 'field %(name)r is readonly', locals())
                    continue
            except AttributeError as e:
                counter[('skipped', name, 'AttributeError', 'getMutator')] += 1
                tell(BASE, 'AttributeError: getMutator')
                continue
            try:
                base_unit = field.getBaseUnit(instance)
            except AttributeError as e:
                counter[('skipped', name, 'AttributeError', 'getBaseUnit')] += 1
                tell(BASE, 'AttributeError: getBaseUnit')
                continue
            tell(DEBUG, 'OK: Feld %(name)r hat Mutator und BaseUnit', locals())
            mimetype = base_unit.mimetype
            if (from_mimetypes is not None
                and mimetype not in from_mimetypes
                ):
                tell(SKIP, 'skipping %(mimetype)r field %(name)r', locals())
                counter[('skipped', name, 'type', mimetype)] += 1
                continue
            args = []
            kwargs = {}
            if not base_unit:
                args.append(field.getDefault(instance))
                tell(CHANGE, 'creating field %(name)r', locals())
                kwargs['_initializing_'] = True
            if mimetype != to_mimetype or not base_unit:
                kwargs['mimetype'] = to_mimetype
            if (default_output_type
                    and field.default_output_type != default_output_type):
                kwargs['default_output_type'] = default_output_type
            if args or kwargs:
                kwargs.update({'field': name,
                               })
                if not args:
                    tell(DEBUG, 'Getting existing value for %(name)r ...', locals())
                    oldval = base_unit.getRaw()
                    args.append(oldval)
                    tell(CHANGE, 'using raw value %(oldval)r for field %(name)r', locals())
                tell(CHANGE, 'rectifying field %(name)r', locals())
                counter[('change', name)] += 1
                mapply(mutator, *args, **kwargs)
                changed = True
            else:
                tell(NOP, 'checked field %(name)r', locals())

        if changed:
            counter['changed'] += 1
            if verbose:
                tell(BASE, 'fixing types for %(iinfo)s: DONE', locals())
        else:
            counter['unchanged'] += 1
            if verbose:
                tell(BASE, 'fixing types for %(iinfo)s: NOTHING TO DO', locals())
        return instance

    fix_types.__doc__ = """
        Korrigiere MIME-Typen für das übergebene Objekt

        from_mimetypes =      %(from_mimetypes)s
        to_mimetype =         %(to_mimetype)r
        default_output_type = %(default_output_type)r
        """ % locals()
    return fix_types


def make_fields_inspector(**kwargs):
    """
    Erzeuge eine Funktion, die Informationen über bestimmte Felder ausgibt
    """
    # --------------------------- [ Argumente für die Ausführung ... [
    verbose = kwargs.pop('verbose', True)
    logger = kwargs.pop('logger', None)
    if verbose is None:
        verbose = logger is not None
    tell = make_minilogger(verbose, logger, 'inspect')

    if 'counter' in kwargs:
        counter = kwargs.pop('counter')
    else:
        counter = defaultdict(int)

    # zu inspizierende Attribute:
    fieldnames_good = makeSet(kwargs.pop('fieldnames_good', None))
    fieldnames_broken = makeSet(kwargs.pop('fieldnames_broken', None))
    skip = make_skip_function(tell,
                              sets=fieldnames_good+fieldnames_broken,
                              **kwargs)
    # --------------------------- ] ... Argumente für die Ausführung ]
    def this_cat(name):
        if name in fieldnames_good:
            return 2
        if name in fieldnames_broken:
            return 1
        return 0

    cat_label = ['Attribute ohne Einordnung:',
                 'Negativ-Beispiele:',
                 'Positiv-Beispiele:',
                 ]
    field_attributes = ['metatype',
                        'mimetype',
                        'default_output_type',
                        ]
    baseunit_attributes = ['mimetype',
                           ]

    def inspect_field(field, instance):
        """
        Funktion in Entwicklung ...
        """
        tell(BASE, '* Feld %r (%s):', field.__name__, field)
        has_attrs = set()
        lacks_attrs = set()
        field_value = {}
        for attr_name in field_attributes:
            try:
                attr = getattr(field, attr_name)
                has_attrs.add(attr_name)
                field_value[attr_name] = attr
                tell(BASE, '  %(attr_name)s: %(attr)r', locals())
            except AttributeError:
                lacks_attrs.add(attr_name)
        try:
            base_unit = field.getBaseUnit(instance)
            tell(BASE, '* %(base_unit)r:', locals())
        except AttributeError as e:
            tell(BASE, '  (keine BaseUnit)')
        else:
            for attr_name in baseunit_attributes:
                try:
                    attr = getattr(base_unit, attr_name)
                    if attr_name in has_attrs:
                        if attr == field_value[attr_name]:
                            suffix = ' (gleich)'
                        else:
                            suffix = ' (UNGLEICH)'
                    else:
                        suffix = ''
                    tell(BASE, '  - %(attr_name)s: %(attr)r%(suffix)s', locals())
                except AttributeError:
                    pass
            raw = base_unit.getRaw()
            tell(BASE, '  = %r', raw[:30])

    def inspect(instance=None, brain=None):
        instance, iinfo = instance_and_label(instance, brain, verbose)
        schema = instance.Schema()
        prev_cat = None
        for field in schema.values():
            name = field.__name__
            if skip(name):
                continue
            if prev_cat is None:
                tell(iinfo)
            curr_cat = this_cat(name)
            if curr_cat != prev_cat:
                tell(cat_label[curr_cat])
                prev_cat = curr_cat

            inspect_field(field, instance)
